import os
import json
from typing import Dict
from pathlib import Path
from agilemind.tool import get_all_tools
from agilemind.prompt import fixed_prompt
from agilemind.execution import Agent, Runner

quality_assurance = Agent(
    "quality_assurance",
    "assure software quality",
    fixed_prompt.QUALITY_ASSURANCE_PROMPT,
)
programmer = Agent(
    "programmer",
    "implement software",
    fixed_prompt.PROGRAMER_PROMPT,
    tools=get_all_tools(),
    handoffs=[quality_assurance],
)
quality_assurance.next_agent = programmer
architect = Agent(
    "architect",
    "create software architecture",
    fixed_prompt.ARCHITECT_PROMPT,
    next_agent=programmer,  # Force handoff to programmer
)
demand_analyst = Agent(
    "demand_analyst",
    "analyze user demand",
    fixed_prompt.DEMAND_ANALYST_PROMPT,
    next_agent=architect,  # Force handoff to architect
)


def dev(demand: str, output: str, model: str, max_iterations: int) -> Dict[str, str]:
    """
    Run the LLM-Agent workflow pipelines.

    Args:
        demand: User demand for the software
        output: Directory path to save the software
        model: String name of the model to use
        max_iterations: Maximum number of iterations to run

    Returns:
        Dictionary containing the software development process
    """
    Path(output).mkdir(parents=True, exist_ok=True)

    # Change current working directory to the output directory
    initial_cwd = os.getcwd()
    os.chdir(output)

    try:
        runner = Runner()
        result = runner.run(demand_analyst, demand, 5)

        with open("trace.txt", "w") as f:
            f.write(json.dumps(result, indent=4))
    finally:
        os.chdir(initial_cwd)  # Restore original working directory

    return result



from typing import Any, Dict, Optional
from dataclasses import dataclass, field
from agilemind.execution import GenerationParams


@dataclass
class TaskAgent:
    """Agent configuration for a task."""

    # Background information of the task, will be used as system message for the agent
    background: Optional[str]
    # Model to use for this task, if not provided, the default model will be used
    model: Optional[str] = None
    # Whether to allow to use a tool for this task
    use_tool: bool = False
    # Configuration for the agent to use for this task
    config: Optional[GenerationParams] = None


@dataclass
class TaskResult:
    """Result of a task execution."""

    # Whether allowed to use a tool for the task
    use_tool: bool = False
    # Output of the task execution
    output: Optional[str] = None
    # Error message if the task failed
    error: Optional[str] = None


@dataclass
class Task:
    """Base class representing a task that needs to be performed by an agent.

    A task encapsulates the work that needs to be done and tracks the result of its execution.
    """

    # Name of the task
    name: str
    # Input data for the task
    input: str = None
    # Agent configuration for the task
    agent: TaskAgent = field(default_factory=TaskAgent)
    # Result of the task execution
    result: TaskResult = field(default_factory=TaskResult)
    # Status of the task, may be: pending, in_progress, completed, failed
    status: str = field(default="pending", init=False)
    # Whether to save the artifact generated by the task
    save_artifact: bool = False
    # Path to save the artifact generated by the task
    artifact_path: Optional[str] = None

    def __post_init__(self):
        """Validate task attributes after initialization."""
        if not self.name:
            raise ValueError("Task must have a name")
        if not self.input:
            raise ValueError("Task must have input data")

    def set_failed(self, error: Optional[str] = None) -> None:
        """Mark the task as failed with optional error information."""
        self.status = "failed"
        self.result.error = error or "<Task failed for unknown reason>"

    def is_complete(self) -> bool:
        """Check if the task is completed."""
        return self.status == "completed"

    def is_failed(self) -> bool:
        """Check if the task has failed."""
        return self.status == "failed"

    def __str__(self) -> str:
        """String representation of the task."""
        return f"Task({self.name}, status={self.status})"

    @classmethod
    def from_dict(cls, task_dict: Dict[str, Any]) -> "Task":
        """Create a Task instance from a dictionary."""
        return cls(**task_dict)



"""
Stage module for organizing tasks within a pipeline.
"""

from agilemind.task import Task
from agilemind.context import Context
from agilemind.execution import Executor
from dataclasses import dataclass, field


@dataclass
class Stage:
    """A stage in a pipeline, containing one or more tasks to be executed."""

    name: str
    description: str = None
    tasks: list[Task] = field(default_factory=list)

    def __post_init__(self):
        if not self.name:
            raise ValueError("Stage must have a name")

    def add_task(self, *tasks: Task) -> "Stage":
        """
        Add one or more tasks to the stage.

        Args:
            tasks: The task(s) to add

        Returns:
            Self for chaining
        """
        self.tasks.extend(tasks)
        return self

    def execute(self, context: Context) -> Context:
        """
        Execute all tasks in the stage and update the context.

        Args:
            context: Current pipeline context

        Returns:
            Updated context with task results
        """

        executor: Executor = context.get("executor")
        if not executor:
            raise ValueError("No executor found in context")

        for i, task in enumerate(self.tasks):
            try:
                context = executor.execute(task, context)

                # Stop execution if task failed and we're not set to continue
                if task.get_status() == "failed" and not context.get(
                    "continue_on_failure", False
                ):
                    break

            except Exception as e:
                context[f"task_{task.name}_error"] = str(e)
                if not context.get("continue_on_failure", False):
                    raise

        return context



"""
Classes for tracking token usage.
"""

from datetime import datetime
from typing import Dict, List, Any, Optional


class TokenCount:
    """Class to track prompt, completion, and total tokens."""

    def __init__(self):
        """Initialize token counts."""
        self.prompt_tokens = 0
        self.completion_tokens = 0
        self.total_tokens = 0

    def update(self, prompt_tokens: int, completion_tokens: int) -> None:
        """
        Update token counts.

        Args:
            prompt_tokens: Number of prompt tokens to add
            completion_tokens: Number of completion tokens to add
        """
        self.prompt_tokens += prompt_tokens
        self.completion_tokens += completion_tokens
        self.total_tokens = self.prompt_tokens + self.completion_tokens

    def to_dict(self) -> Dict[str, int]:
        """Convert to dictionary representation."""
        return {
            "prompt_tokens": self.prompt_tokens,
            "completion_tokens": self.completion_tokens,
            "total_tokens": self.total_tokens,
        }


class TokenDetail:
    """Class to represent a single token usage record."""

    def __init__(
        self,
        agent_name: str,
        round_number: int,
        prompt_tokens: int,
        completion_tokens: int,
        model: Optional[str] = None,
    ):
        """Initialize token detail record."""
        self.timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        self.agent = agent_name
        self.round = round_number
        self.model = model
        self.prompt_tokens = prompt_tokens
        self.completion_tokens = completion_tokens
        self.total_tokens = prompt_tokens + completion_tokens

    def to_dict(self) -> Dict[str, Any]:
        """Convert to dictionary representation."""
        return {
            "timestamp": self.timestamp,
            "agent": self.agent,
            "round": self.round,
            "model": self.model,
            "prompt_tokens": self.prompt_tokens,
            "completion_tokens": self.completion_tokens,
            "total_tokens": self.total_tokens,
        }


class AgentTokenUsage:
    """Class to track token usage by agent."""

    def __init__(self):
        """Initialize agent token usage tracking."""
        self._agents: Dict[str, TokenCount] = {}

    def update(
        self, agent_name: str, prompt_tokens: int, completion_tokens: int
    ) -> None:
        """
        Update token usage for a specific agent.

        Args:
            agent_name: Name of the agent
            prompt_tokens: Number of prompt tokens to add
            completion_tokens: Number of completion tokens to add
        """
        if agent_name not in self._agents:
            self._agents[agent_name] = TokenCount()

        self._agents[agent_name].update(prompt_tokens, completion_tokens)

    def __getattr__(self, agent_name: str) -> TokenCount:
        """Access agent token usage via attribute."""
        if agent_name not in self._agents:
            self._agents[agent_name] = TokenCount()
        return self._agents[agent_name]

    def to_dict(self) -> Dict[str, Dict[str, int]]:
        """Convert to dictionary representation."""
        return {name: count.to_dict() for name, count in self._agents.items()}


class TokenUsage:
    """Main class for tracking all token usage."""

    def __init__(self):
        """Initialize token usage tracking."""
        self.total = TokenCount()
        self.agent = AgentTokenUsage()
        self.detailed: List[TokenDetail] = []

    def update(
        self,
        prompt_tokens: int,
        completion_tokens: int,
        agent_name: str,
        round_number: int,
        model: Optional[str] = None,
    ) -> None:
        """
        Update all token usage statistics.

        Args:
            prompt_tokens: Number of prompt tokens used
            completion_tokens: Number of completion tokens used
            agent_name: Name of the agent making the call
            round_number: Round number within the agent's processing
            model: Optional model name used for the API call
        """
        # Update total counts
        self.total.update(prompt_tokens, completion_tokens)

        # Update per-agent counts
        self.agent.update(agent_name, prompt_tokens, completion_tokens)

        # Add detailed record
        detail = TokenDetail(
            agent_name=agent_name,
            round_number=round_number,
            prompt_tokens=prompt_tokens,
            completion_tokens=completion_tokens,
            model=model,
        )
        self.detailed.append(detail)

    def to_dict(self) -> Dict[str, Any]:
        """Convert to dictionary representation."""
        return {
            "total": self.total.to_dict(),
            "by_agent": self.agent.to_dict(),
            "detailed": [detail.to_dict() for detail in self.detailed],
        }



"""
Module to store and manage LLM model pricing information.
"""

from enum import Enum
from typing import Dict
from dataclasses import dataclass
from rich import print as rich_print

# Set to track models that have already been warned about
_warned_models = set()


@dataclass
class ModelPrice:
    """Class for storing model pricing information."""

    # Input price in USD per 1000 tokens
    input_price: float
    # Output price in USD per 1000 tokens
    output_price: float
    # Human-readable model name
    name: str


class ModelPricing(Enum):
    """Enum for common LLM models and their pricing."""

    # OpenAI models
    GPT_4O_MINI = ModelPrice(0.00015, 0.0006, "GPT-4o Mini")
    GPT_4O = ModelPrice(0.0025, 0.01, "GPT-4o")

    # Anthropic models
    CLAUDE_3_7_SONNET_20250219 = ModelPrice(
        0.0033, 0.0165, "Claude 3.7 Sonnet 20250219"
    )

    # Deepseek models
    DEEPSEEK_V3 = ModelPrice(0.000272, 0.001088, "DeepSeek V3")
    DEEPSEEK_R1 = ModelPrice(0.000546, 0.002184, "DeepSeek R1")

    def known_models() -> list:
        """Return a list of known model names"""
        return [model.name for model in ModelPricing]


def calculate_cost(
    model: str, prompt_tokens: int, completion_tokens: int
) -> Dict[str, float]:
    """
    Calculate the cost for a given model and token usage.

    Args:
        model: The model to calculate pricing for
        prompt_tokens: Number of input tokens
        completion_tokens: Number of output tokens

    Returns:
        Dictionary with prompt_cost, completion_cost, and total_cost
    """
    # Format model name for comparison
    model = model.upper().replace("-", "_").replace(" ", "_")

    if model not in ModelPricing.known_models():
        if model not in _warned_models:
            rich_print(
                f"[yellow]Unknown price for model: {model}, using GPT-4o's price for reference.[/yellow]"
            )
            _warned_models.add(model)
        model = "GPT_4O"

    model_price = ModelPricing[model.upper()].value
    prompt_cost = model_price.input_price * (prompt_tokens / 1000)
    completion_cost = model_price.output_price * (completion_tokens / 1000)
    total_cost = prompt_cost + completion_cost

    return {
        "prompt_cost": prompt_cost,
        "completion_cost": completion_cost,
        "total_cost": total_cost,
    }



import base64


def generate_image_message(image_path: str, str_message: str) -> str:
    """
    Generate a message with an image.

    Args:
        image_path (str): The path to the image.
        str_message (str): The message to be sent.

    Returns:
        str: The message of OpenAI API format, with base64 encoded image.
    """
    with open(image_path, "rb") as f:
        image = f.read()
    base64_image = base64.b64encode(image).decode("utf-8")
    base64_image = f"data:image/jpeg;base64,{base64_image}"

    message = {
        "role": "user",
        "content": [
            {"type": "text", "text": str_message},
            {"type": "image_url", "image_url": {"url": base64_image}},
        ],
    }

    return message



import os
import sys
import ast
import importlib
from .interface import AbsChecker
from typing import Dict, List, Tuple


def extract_imports(code: str) -> List[Tuple[str, str, List[str], int]]:
    """
    Extract all import statements from Python code string.

    Args:
        code (str): Python code as string

    Returns:
        out (list of tuples):
            - import type ('import' or 'from')
            - module name
            - list of names being imported (empty for simple imports)
            - line number in the source code
    """
    tree = ast.parse(code)
    imports = []

    for node in ast.walk(tree):
        if isinstance(node, ast.Import):
            for name in node.names:
                imports.append(("import", name.name, [], node.lineno))
        elif isinstance(node, ast.ImportFrom):
            module = node.module or ""
            names = [alias.name for alias in node.names]
            imports.append(("from", module, names, node.lineno))

    return imports


def check_imports(code: str, local_dir: str = None) -> Dict[Tuple[str, int], str]:
    """
    Check if all imports in the code can be successfully imported.

    Args:
        code: Python code as string
        local_dir: Directory to check for local modules (defaults to current working directory)

    Returns:
        Dictionary mapping (import statement, line number) to error messages (empty if successful)
    """
    imports = extract_imports(code)
    results = {}

    # If no local_dir specified, use current working directory
    if local_dir is None:
        local_dir = os.getcwd()

    # Temporarily add local_dir to sys.path to find local modules
    original_path = sys.path.copy()
    if local_dir not in sys.path:
        sys.path.insert(0, local_dir)

    try:
        for imp_type, module_name, names, lineno in imports:
            if imp_type == "import":
                import_stmt = f"import {module_name}"
                try:
                    importlib.import_module(module_name)
                except Exception as e:
                    results[(import_stmt, lineno)] = str(e)
            elif imp_type == "from":
                if not names:
                    continue

                import_stmt = f"from {module_name} import {', '.join(names)}"
                try:
                    if module_name:
                        mod = importlib.import_module(module_name)
                        for name in names:
                            try:
                                getattr(mod, name)
                            except AttributeError:
                                results[(import_stmt, lineno)] = (
                                    f"Cannot import name '{name}' from '{module_name}'"
                                )
                except Exception as e:
                    results[(import_stmt, lineno)] = str(e)
    finally:
        # Restore the original sys.path
        sys.path = original_path

    return results


def format_error_message(errors: Dict[Tuple[str, int], str]) -> str:
    """
    Format the error message for display including line numbers of invalid imports.

    Args:
        errors: Dictionary of (import statements, line number) to error messages

    Returns:
        Formatted error message
    """
    if not errors:
        return "All imports are successful!"

    error_msg = "Errors in imports:"
    for (stmt, lineno), error in errors.items():
        error_msg += f'\n  - Line {lineno}: "{stmt}": {error}'

    return error_msg


class ImportChecker(AbsChecker):
    """
    A checker that verifies if all imports in the code can be successfully imported.
    """

    @property
    def name(self) -> str:
        return "Import Checker"

    def check(self, code: str) -> List[Dict[str, str]]:
        errors = check_imports(code)
        return [
            {"line": lineno, "column": 0, "problematic_code": stmt, "message": error}
            for (stmt, lineno), error in errors.items()
        ]



"""
Runner module for managing the execution flow between agents.
"""

import json
from .agent import Agent
from typing import Dict, Any


class Runner:
    """
    Manages the execution flow between agents.

    The runner starts with an initial agent and input, then handles
    any handoffs between agents, tool usage, and final output generation.
    """

    @classmethod
    def run(
        cls, starting_agent: Agent, input: str, max_iterations: int = 5
    ) -> Dict[str, Any]:
        """
        Run a conversation through agents, handling handoffs.

        Args:
            starting_agent: The agent to start with
            input: Initial user input
            max_iterations: Maximum number of agent handoffs to prevent infinite loops

        Returns:
            Dict containing the final response and execution trace
        """
        current_agent = starting_agent
        current_input = input
        iterations = 0
        execution_trace = []

        while iterations < max_iterations:
            iterations += 1

            # Process the input with the current agent
            result = current_agent.process(current_input)

            # Add to execution trace
            execution_trace.append(
                {"agent": current_agent.name, "input": current_input, "output": result}
            )

            # Check if there's a handoff (could be from direct handoff or forced via next_agent)
            if result["handoff"]:
                # Update the trace to indicate the handoff
                execution_trace[-1]["handoff_to"] = result["handoff"].name
                # Update current agent and use the previous agent's response as input for the new agent
                current_agent = result["handoff"]
                # Use previous agent's response as new input
                current_input = result["content"]
                # Reset the handoff to the new agent's name
                result["handoff"] = result["handoff"].name
                continue

            if result["tool_calls"]:
                # Add to the trace
                execution_trace[-1]["tool_calls"] = result["tool_calls"]

                # Format the tool results into a message to send back to the agent
                current_input = f'Here are the results of executed tools:\n{json.dumps(result["tool_calls"])}\n\nPlease continue based on these results.'
                continue

            # No handoffs or tool follow-ups needed, we're done
            return {
                "agent": current_agent.name,
                "response": result["content"],
                "execution_trace": execution_trace,
            }

        # If we reach here, we hit the max iterations
        return {
            "agent": current_agent.name,
            "response": "Maximum number of agent iterations reached without resolution.",
            "execution_trace": execution_trace,
        }



import inspect
import functools
from dataclasses import dataclass
from typing import Any, Callable, Optional, Type, Union, get_type_hints


@dataclass
class ParameterSpec:
    """Specification for a parameter in a tool function"""

    name: str
    type: Type
    description: str
    required: bool = True
    default: Any = None


def tool(
    name_or_func: Optional[Union[str, Callable]] = None,
    description: Optional[str] = None,
    confirmation_required: bool = False,
    group: str = "general",
):
    """
    Decorator to mark a function as a tool and provide OpenAI API tools format metadata.

    Args:
        name_or_func: Function or custom name for the tool
        description: Optional description (defaults to function docstring)
        confirmation_required: Whether user confirmation is required before execution
        group: Group/category this tool belongs to (default: "general")

    Returns:
        Decorated function with OpenAI tools metadata
    """
    # Handle case when decorator is used without parentheses @tool
    if callable(name_or_func):
        return _tool_impl(name_or_func)

    # Handle case when decorator is used with parentheses @tool(...) or @tool()
    name = name_or_func

    def decorator(func: Callable):
        return _tool_impl(
            func,
            name=name,
            description=description,
            confirmation_required=confirmation_required,
            group=group,
        )

    return decorator


def _tool_impl(
    func: Callable,
    name: Optional[str] = None,
    description: Optional[str] = None,
    confirmation_required: bool = False,
    group: str = "general",
):
    """
    Actual implementation of the tool decorator

    Args:
        func: The function to decorate
        name: Optional custom name for the tool
        description: Optional description
        confirmation_required: Whether user confirmation is required before execution
        group: Group/category this tool belongs to
    """
    func_name = name or func.__name__
    signature = inspect.signature(func)
    doc = inspect.getdoc(func) or ""
    func_description = description or (
        doc.split("\n\n")[0] if doc else f"Function {func_name}"
    )

    # Extract parameter specifications
    params = []
    for param_name, param in signature.parameters.items():
        if param_name == "self":
            continue

        is_required = param.default is inspect.Parameter.empty
        default_value = None if is_required else param.default

        # Try to find parameter description in docstring
        param_desc = ""
        if doc:
            doc_lines = doc.split("\n")
            for i, line in enumerate(doc_lines):
                if line.strip().startswith(f"{param_name}:"):
                    param_desc = line.split(":", 1)[1].strip()
                # Look for Args section format
                elif "Args:" in doc and line.strip().startswith(f"{param_name}:"):
                    param_desc = line.split(":", 1)[1].strip()
                # Look for parameters documented in the format param_name (type): description
                elif f"{param_name} (" in line:
                    param_desc = line.split(":", 1)[1].strip() if ":" in line else ""

        param_type = get_type_hints(func).get(param_name, str)

        params.append(
            ParameterSpec(
                name=param_name,
                type=param_type,
                description=param_desc,
                required=is_required,
                default=default_value,
            )
        )

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    # Attach OpenAI tools metadata
    wrapper.is_tool = True
    wrapper.tool_name = func_name
    wrapper.tool_description = func_description
    wrapper.tool_parameters = params
    wrapper.confirmation_required = confirmation_required
    wrapper.tool_group = group

    # Add a method to get OpenAI format
    def get_openai_schema():
        properties = {}
        required_params = []

        for param in params:
            # Skip the context parameter when generating schema for OpenAI API
            if param.name == "context":
                continue

            param_schema = {
                "type": _get_openai_type(param.type),
                "description": param.description,
            }

            if not param.required:
                param_schema["default"] = param.default
            else:
                required_params.append(param.name)

            properties[param.name] = param_schema

        schema = {
            "type": "function",
            "function": {
                "name": func_name,
                "description": func_description,
                "parameters": {
                    "type": "object",
                    "properties": properties,
                    "required": required_params,
                },
            },
        }

        # Add confirmation metadata if required
        if confirmation_required:
            schema["function"]["confirmation_required"] = True

        return schema

    wrapper.get_openai_schema = get_openai_schema
    return wrapper


def _get_openai_type(python_type):
    """Convert Python types to OpenAI schema types"""
    if python_type == str:
        return "string"
    elif python_type == int:
        return "integer"
    elif python_type == float:
        return "number"
    elif python_type == bool:
        return "boolean"
    elif hasattr(python_type, "__origin__") and python_type.__origin__ == list:
        return "array"
    elif hasattr(python_type, "__origin__") and python_type.__origin__ == dict:
        return "object"
    else:
        return "string"  # Default fallback



"""
Configuration module for the Executor class.
"""

import os
from dotenv import load_dotenv
from typing import Any, Dict, List, Optional
from dataclasses import asdict, dataclass, field


@dataclass
class GenerationParams:
    """
    Dataclass to represent the generation parameters for the Executor class.
    None values represent parameters that are default to the API.
    """

    top_p: Optional[float] = None
    max_tokens: Optional[int] = None
    max_completion_tokens: Optional[int] = None
    tools: Optional[List[Dict]] = None
    temperature: Optional[float] = None

    def __post_init__(self):
        if self.max_tokens is not None:
            self.max_completion_tokens = self.max_tokens
        if self.max_completion_tokens is not None:
            self.max_tokens = self.max_completion_tokens

    def to_dict(self) -> Dict[str, Any]:
        """Convert to dictionary, excluding None values."""
        return {k: v for k, v in asdict(self).items() if v is not None}


@dataclass
class ExecutorConfig:
    """
    Dataclass to represent the configuration parameters for the Executor class.
    """

    api_key: str
    default_model: str
    base_url: Optional[str] = None
    generation_params: GenerationParams = field(default_factory=GenerationParams)

    def to_dict(self) -> Dict[str, Any]:
        """Convert configuration to dictionary for serialization."""
        result = {
            "api_key": "REDACTED" if self.api_key else None,
            "default_model": self.default_model,
            "base_url": self.base_url,
            "generation_params": self.generation_params.to_dict(),
        }
        return result

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "ExecutorConfig":
        """Create configuration from dictionary."""
        api_key = os.getenv("OPENAI_API_KEY")
        if not api_key:
            raise ValueError("API key is required, please set OPENAI_API_KEY")

        return cls(
            api_key=api_key,
            default_model=data["default_model"],
            base_url=data["base_url"],
            generation_params=GenerationParams(**data["generation_params"]),
        )

    @classmethod
    def from_env(cls) -> "ExecutorConfig":
        """Create configuration from environment variables."""
        load_dotenv()

        api_key = os.getenv("OPENAI_API_KEY")
        if not api_key:
            raise ValueError("API key is required, please set OPENAI_API_KEY")

        base_url = os.getenv("OPENAI_BASE_URL")
        default_model = os.getenv("AM_DEFAULT_MODEL")

        max_tokens = os.getenv("AM_MAX_TOKENS")
        top_p = os.getenv("AM_TOP_P")
        temperature = os.getenv("AM_DEFAULT_TEMPERATURE")

        generation_params = GenerationParams(
            max_tokens=int(max_tokens) if max_tokens else None,
            top_p=float(top_p) if top_p else None,
            temperature=float(temperature) if temperature else None,
        )

        return cls(
            api_key=api_key,
            default_model=default_model,
            base_url=base_url,
            generation_params=generation_params,
        )



"""
Development of software using agile methodology.
"""

import os
import json
import shutil
import readchar
from pathlib import Path
from rich.panel import Panel
from rich.align import Align
from rich.console import Console
from rich import print as rprint
from agilemind.tool import get_tool
from agilemind.execution import Agent
from agilemind.context import Context
from typing import Dict, List, Optional
from agilemind.prompt import agile_prompt
from concurrent.futures import ThreadPoolExecutor, as_completed
from agilemind.utils import (
    LogWindow,
    load_config,
    create_file_tree,
    extract_agent_llm_config,
    convert_json_to_markdown,
)

config = load_config()
console = Console()

prototype_builder = Agent(
    name="prototype_builder",
    description="Build prototype of the software",
    instructions=agile_prompt.PROTOTYPE_DEVELOPER,
    tools=[get_tool("write_file")],
    **extract_agent_llm_config("prototype", config),
)
architect = Agent(
    name="architect",
    description="Design architecture of the software",
    instructions=agile_prompt.ARCHITECT,
    tools=[get_tool("write_file")],
    **extract_agent_llm_config("architecture", config),
)
developer = Agent(
    name="developer",
    description="Implement code for the software",
    instructions=agile_prompt.DEVELOPER,
    tools=[get_tool("write_file")],
    **extract_agent_llm_config("programming", config),
)

all_agents = [prototype_builder, architect, developer]


def build_prototype(
    context: Context,
    window: LogWindow,
    demand: str,
    max_iterations: int = 5,
) -> tuple["str", "str"]:
    """
    Build a prototype of the software.

    Args:
        context (Context): Context object containing the software development process
        window (LogWindow): CLI window for displaying progress
        demand (str): User demand for the software
        max_iterations (int): Maximum number of iterations to run

    Returns:
        out: Tuple of feedback and prototype
    """
    window.log("Developing prototype of the software...")

    prototype_task = window.add_task("Developing prototype", status="running")

    prototype_builder.process(context, demand, max_iterations)

    if not os.path.isfile("docs/prototype.html"):
        print("Critical: Prototype file not found")
        raise FileNotFoundError("Prototype file not found")
    with open("docs/prototype.html", "r") as f:
        prototype = f.read()

    window.update_task(prototype_task, status="pending")

    client_satisfied = False
    revision_count = 0
    feedback = ""
    while not client_satisfied and revision_count < max_iterations:
        window.hide()
        console.print(
            Panel(
                Align.center(
                    "The prototype has been developed. Please check the prototype and provide feedback. Are you satisfied with the prototype? (Y/n)"
                ),
                border_style="bold blue",
                title="Client Feedback",
            )
        )
        client_satisfied = readchar.readchar().lower() == "y"
        console.clear()

        if not client_satisfied:
            revision_count += 1
            previous_prototype = prototype
            feedback_template = (
                "Given client's demand: \n{demand}\n\n"
                "Previously the prototype is: \n{previous_prototype}\n\n"
                "The client has provided the following feedback for the prototype: \n{feedback}"
            )
            input_text = input("Please provide your feedback for the prototype: ")
            feedback += input_text + "\n"
            feedback_info = feedback_template.format(
                demand=demand,
                previous_prototype=previous_prototype,
                feedback=feedback,
            )

            window.show()
            window.update_task(prototype_task, status="running")
            prototype_builder.process(context, feedback_info, max_iterations)
            window.update_task(prototype_task, status="pending")

            with open("docs/prototype.html", "r") as f:
                prototype = f.read()

    window.show()
    window.complete_task(prototype_task)

    return feedback, prototype


def build_architecture(
    context: Context,
    window: LogWindow,
    demand: str,
    feedback: str,
    prototype: str,
    max_iterations: int = 5,
) -> tuple[List, str]:
    """
    Build a prototype of the software.

    Args:
        context (Context): Context object containing the software development process
        window (LogWindow): CLI window for displaying progress
        demand (str): User demand for the software
        feedback (str): Feedback from the client
        prototype (str): Final prototype of the software
        max_iterations (int): Maximum number of iterations to run

    Returns:
        out: Tuple of code file list and architecture information
    """
    window.log("Designing architecture of the software...")

    architecture_task = window.add_task("Developing architecture", status="running")
    demand_info = agile_prompt.FEEDBACK_TEMPLATE.format(
        raw_demand=demand, feedback=feedback, prototype=prototype
    )

    architect.process(context, demand_info, max_iterations)

    if not os.path.isfile("logs/architecture.json"):
        print("Critical: Architecture file not found")
        raise FileNotFoundError("Architecture file not found")
    with open("logs/architecture.json", "r") as f:
        json_info: Dict = json.load(f)

    md_info = json_info.copy()
    code_file_list_md = create_file_tree(json_info["code_file_list"])
    md_info["code_file_list"] = code_file_list_md
    architecture_md = convert_json_to_markdown(
        "introduction",
        "code_file_list",
        "class_structure",
        "call_flow",
        data=md_info,
        title="Software System Design",
        code_languages={
            "code_file_list": "plaintext",
            "class_structure": "mermaid",
            "call_flow": "mermaid",
        },
    )
    with open("docs/architecture.md", "w") as f:
        f.write(architecture_md)

    window.complete_task(architecture_task)
    return json_info.get("code_file_list", []), architecture_md


def implement_code(
    context: Context,
    window: LogWindow,
    code_file_list: List[str],
    architecture: str,
    max_iterations: int = 5,
) -> None:
    """
    Implement the code for the software.

    Args:
        context (Context): Context object containing the software development process
        window (LogWindow): CLI window for displaying progress
        code_file_list (List[str]): List of code files to implement
        architecture (str): Architecture information of the software
        max_iterations (int): Maximum number of iterations to run

    Returns:
        None
    """
    window.log("Implementing code for the software...")

    code_task = window.add_task("Implementing code", status="running")

    with ThreadPoolExecutor() as executor:
        code_tasks = [
            executor.submit(
                developer.process,
                context,
                agile_prompt.DEVELOPING_TEMPLATE.format(
                    architecture=architecture, file_path=file
                ),
                max_iterations,
            )
            for file in code_file_list
        ]
        for task in as_completed(code_tasks):
            task.result()

    window.log("Code implementation completed.")
    window.complete_task(code_task)

    return


def run_workflow(
    demand: str,
    max_iterations: int = 5,
    model: Optional[str] = None,
) -> dict:
    """
    Run the LLM-Agent workflow pipelines.

    Args:
        demand (str): User demand for the software
        max_iterations (int): Maximum number of iterations to run
        model (str, Optional): String name of the model to use

    Returns:
        out: Dictionary containing the software development process
    """
    if model:
        for agent in all_agents:
            agent.set_model(model)

    output_dir = os.path.abspath(os.getcwd())
    context = Context(demand, output_dir)

    window = LogWindow(title="AgileMind Development")
    window.open()

    window.log("Starting the software development process...")

    feedback, prototype = build_prototype(context, window, demand, max_iterations)
    file_list, architecture = build_architecture(
        context, window, demand, feedback, prototype, max_iterations
    )

    implement_code(context, window, file_list, architecture, max_iterations)

    window.log("Software development process completed. Exiting...")
    window.close()

    return context.dump()


def dev(
    demand: str, output: str, model: Optional[str] = None, max_iterations: int = 5
) -> dict:
    """
    Run the LLM-Agent workflow pipelines.

    Args:
        demand (str): User demand for the software
        output (str): Directory path to save the software
        model (str, Optional): String name of the model to use
        max_iterations (int, Optional): Maximum number of iterations to run

    Returns:
        out: Dictionary containing the software development process
    """
    # If output dir exists, ask user whether to confirm purging it first
    if Path(output).exists():
        rprint(
            Panel(
                Align.center(
                    f'The output directory "{output}" already exists. Do you want to delete its contents? (Y/n)'
                ),
                border_style="bold red",
                title="Warning",
            )
        )

        confirm = readchar.readchar().lower()
        console.clear()

        if confirm != "y":
            return {"status": "cancelled"}

        # Purge the output directory
        shutil.rmtree(output)

    Path(output).mkdir(parents=True, exist_ok=True)
    Path(output, "docs").mkdir(parents=True, exist_ok=True)
    Path(output, "logs").mkdir(parents=True, exist_ok=True)

    # Change current working directory to the output directory
    initial_cwd = os.getcwd()
    os.chdir(output)

    try:
        result = run_workflow(demand, model=model, max_iterations=max_iterations)

        with open("logs/development_record.json", "w") as f:
            f.write(json.dumps(result, indent=4))
    finally:
        os.chdir(initial_cwd)  # Restore original working directory

    return result



import os
import json
import shutil
import subprocess
from .tool_decorator import tool
from agilemind.context import Context
from typing import Any, Dict, Optional


class Tools:
    @staticmethod
    @tool(
        "write_file",
        description="Write content to the file at the specified path. If the file already exists, it will be overwritten. Otherwise, a new file (and any necessary directories) will be created.",
        group="file_system",
    )
    def write_file(context: Context, path: str, content: str) -> Dict[str, Any]:
        """
        Write content to a file. If the file already exists, it will be overwritten. Otherwise, a new file will be created.

        Args:
            path: The path to the file to write. You are currently in the root directory of the project. Use relative path.
            content: The content to write to the file. When creating a code file, make sure it is a valid code.

        Returns:
            Dict containing success status and message
        """
        if ".." in path or path.startswith("/"):
            return {
                "success": False,
                "message": "Cannot write files outside the current directory",
            }

        overwritten = True if os.path.isfile(path) else False

        if (
            not path.endswith(".json")
            and not path.endswith(".md")
            and not path.endswith(".txt")
        ):
            if path not in context.code.structure:
                context.code.structure[path] = content
            context.code.uptodated[path] = content

        try:
            # Ensure directory exists
            os.makedirs(os.path.dirname(os.path.abspath(path)), exist_ok=True)

            with open(path, "w") as f:
                if path.endswith(".json"):
                    json.dump(json.loads(content), f, indent=4)
                else:
                    f.write(content)
            return {
                "success": True,
                "message": (
                    f"File created at {path}"
                    if not overwritten
                    else f"File overwritten at {path}"
                ),
            }
        except Exception as e:
            return {"success": False, "message": f"Failed to create file: {str(e)}"}

    @staticmethod
    @tool(
        "read_file",
        description="Read the content of a file",
        group="file_system",
    )
    def read_file(context: Context, path: str) -> Dict[str, Any]:
        """
        Read and return the content of a file.

        Args:
            path: The path to the file to read. **MUST use relative path.**

        Returns:
            Dict containing success status, message, and file content
        """
        if ".." in path or path.startswith("/"):
            return {
                "success": False,
                "message": "Cannot read files outside the current directory",
            }

        try:
            if not os.path.exists(path):
                return {"success": False, "message": f"File not found: {path}"}

            # Use the correct encoding for reading Chinese characters
            with open(path, "r", encoding="utf-8") as f:
                content = f.read()
            return {
                "success": True,
                "message": f"File read successfully",
                "content": content,
            }
        except Exception as e:
            return {"success": False, "message": f"Failed to read file: {str(e)}"}

    @staticmethod
    @tool(
        "execute_command",
        description="Execute a shell command",
        confirmation_required=True,
    )
    def execute_command(
        context: Context, command: str, cwd: Optional[str] = None
    ) -> Dict[str, Any]:
        """
        Execute a shell command. Related path **MUST be relative path.**

        Args:
            command: The command to execute
            cwd: Current working directory (optional)

        Returns:
            Dict containing success status, message, stdout and stderr
        """
        try:
            result = subprocess.run(
                command, shell=True, cwd=cwd, capture_output=True, text=True
            )

            return {
                "success": result.returncode == 0,
                "message": f"Command executed with return code {result.returncode}",
                "stdout": result.stdout,
                "stderr": result.stderr,
                "returncode": result.returncode,
            }
        except Exception as e:
            return {"success": False, "message": f"Failed to execute command: {str(e)}"}

    @staticmethod
    @tool(
        "list_project_structure",
        description="List the structure of the project",
        group="file_system",
    )
    def list_directory(context: Context) -> Dict[str, Any]:
        """
        List the structure of the project.

        Returns:
            Dict containing success status, message, and project structure
        """
        try:
            items = list(context.code.uptodated.keys())
            return {
                "success": True,
                "message": "Project structure listed",
                "items": items,
            }
        except Exception as e:
            return {
                "success": False,
                "message": f"Failed to list project structure: {str(e)}",
            }

    @staticmethod
    @tool(
        "delete_file",
        description="Delete a file or directory",
    )
    def delete_file(context: Context, path: str) -> Dict[str, Any]:
        """
        Delete a file or directory.

        Args:
            path: The path of file to delete. **MUST use relative path.**

        Returns:
            Dict containing success status and message
        """
        if ".." in path or path.startswith("/"):
            return {
                "success": False,
                "message": "Cannot delete files outside the current directory",
            }

        try:
            if not os.path.exists(path):
                return {"success": False, "message": f"Path not found: {path}"}

            if os.path.isdir(path):
                shutil.rmtree(path)
                return {"success": True, "message": f"Directory deleted: {path}"}
            else:
                os.remove(path)
                return {"success": True, "message": f"File deleted: {path}"}
        except Exception as e:
            return {"success": False, "message": f"Failed to delete: {str(e)}"}

    @staticmethod
    @tool(
        "add_to_requirements",
        description="Add a package to the requirements file",
        group="development",
    )
    def add_to_requirements(
        context: Context, language: str, package_name: str, version: str = None
    ) -> Dict[str, Any]:
        """
        Add a package to the requirements file based on the language

        Args:
            language: The language for which to add the package
            package_name: The name of the package to add
            version: The version of the package to add (optional)

        Returns:
            Dict containing success status and message
        """
        if language.lower() == "python":
            with open("requirements.txt", "a") as f:
                if version:
                    f.write(f"{package_name}=={version}\n")
                else:
                    f.write(f"{package_name}\n")
            return {
                "success": True,
                "message": f"Added {package_name} to requirements.txt",
            }
        elif language.lower() == "javascript":
            if not os.path.exists("package.json"):
                with open("package.json", "w") as f:
                    data = {"dependencies": {}}
                    json.dump(data, f, indent=2)
            else:
                with open("package.json", "r") as f:
                    data = json.load(f)
            if "dependencies" not in data:
                data["dependencies"] = {}
            data["dependencies"][package_name] = version or "*"
            with open("package.json", "w") as f:
                json.dump(data, f, indent=2)
            return {"success": True, "message": f"Added {package_name} to package.json"}

        return {
            "success": False,
            "message": f"Unsupported language: {language}",
        }

    @staticmethod
    @tool(
        "get_project_structure",
        description="Get the code structure of a module or all modules",
        group="development",
    )
    def get_code_structure(context: Context, module: str = None) -> Dict[str, Any]:
        """
        Get the code structure of a module or all modules

        Args:
            module: The name of the module to get the structure of (optional)

        Returns:
            Dict containing success status and message
        """
        try:
            code_structure = {
                k: v
                for k, v in context.code.structure.items()
                if module is None or module in k
            }

            return {
                "success": True,
                "message": f"Code structure of {module or 'all modules'} retrieved",
                "code_structure": code_structure,
            }
        except Exception as e:
            return {
                "success": False,
                "message": f"Failed to get code structure: {str(e)}",
            }



"""Prompt texts for development using Agile model."""

PROTOTYPE_DEVELOPER = """
You are an expert full-stack developer from an software development team called "Agile Mind". You excel in product development and UI/UX design. This team follows the agile model to develop software.

Your job is to develop a prototype based on the client's requirements, which will be shown to the client for feedback.

Follow these steps:
1. Read and understand the requirements carefully, consider what the client wants and needs.
2. From a project management perspective, plan the functionalities, UI/UX design of the software.
3. From a UI/UX design perspective, design the software interface.
4. Generate all the prototype views to a HTML file, whose path is **"docs/prototype.html"**. You may use FontAwesome and TailwindCSS for icons and styles instead of plain HTML/CSS. Your goal is to make the prototype look as real as possible, so the client can confirm the design.

Note that:
- The prototype use HTML just to show its views. It does not mean the final software will be developed using HTML. Ignore client's demand for the programming language, platform, etc. You should explain this at footer of the page.
- The HTML should list all the views of the software, so the client can understand the software's functionalities.

Use "write_file" tool to generate the HTML file.
"""


PROJECT_MANAGER = """
You are an expert project manager from an software development team called "Agile Mind". This team follows the agile model to develop software. "Agile Mind" team consists of several groups of developers. 

Givem the client's demand, feedback from the UI design, and the final UI design, your job is to manage the project and plan the development process, by dividing the project into several tasks and assigning them to different groups. Each task will be assigned to a group of developers.

Follow these steps:
1. Read and understand the demand, feedback, and the final UI design carefully.
2. Judge the project's complexity and decide how many tasks should be divided. The software repo may contain only one file or multiple modules.
3. Decompose the project into several feasible tasks. Use the handoff tool to instruct the developer team to develop the software repository based on the client's requirements. Calling multiple times means multiple tasks for different groups.

Note that:
- Start from scratch. The UI-design HTML is ONLY used as reference for UI/UX design. Do not use it as a base for development!
- If not specified, the client want an application instead of a frontend/backend project. Do NOT tend to develop frontend/backend code if not specified.
- Each task should be clear and concise. Make sure the developers understand what they need to do.
- Do not design CI/CD, testing, etc. Only focus on the development process.
- You should try to reduce the dependencies between tasks, so that the development process can be parallelized.
- Make sure your instructions are clear and concise, so that the developers can understand what they need to do. Instruct them to follow the client's requirements, use the provided tools to write files, and start from scratch. Your goal is to make the development process as smooth as possible.
"""

ARCHITECT = """
You are an expert architect from an software development team called "Agile Mind". This team follows the agile model to develop software.

Givem the client's demand, feedback from the UI design, and the final UI design, your job is to design the software architecture.

Follow these steps:
1. Read and understand the demand, feedback, and the final UI design carefully.
2. Design the software architecture based on the client's requirements. 
3. Output in JSON format with path **"logs/architecture.json"**, containing the following field:
    - introduction: a brief introduction to the software.
    - code_file_list: the file **path** (e.g. "src/core/logic.py" for a python app) **list** of the software repository. There should be an entry point file located in the root directory (e.g. "main.py" for a Python app). Valid JSON list.
    - class_structure: the class structure, valid **Mermaid** class diagram with all the classes, their properties and methods, especially the parameters of constructors.
    - call_flow: the call flow of the software, valid **Mermaid** sequence diagram. Use method names and all parameters in the diagram.

Note that:
- "code_file_list" should only list the code files. Do not list other files such as documents, assets, etc.
- The diagrams should be clear and concise enough, so that the developers can understand the software architecture quickly and correctly.
- The methods in class_structure should show its name and **parameters**, e.g. "method_name(param1: type1, param2: type2): return_type".
- The name of class, method, and parameters should follow the language's naming convention, e.g. PEP8 for Python.

Use "write_file" tool to generate the JSON file.
"""

DEVELOPER = """
You are an expert developer from an software development team called "Agile Mind". This team follows the agile model to develop software.

Given the architecture design and a file path from the file list, your job is to develop this specific file of the software repository.

Follow these steps:
1. Read and understand the architecture design carefully.
2. Develop the file based on the architecture design. 

Note that:
- Use the provided tools to write files.
- Stick to the path and file name provided in the architecture design.
- Stick to the class structure (class name, method name, properties, etc.) and call flow provided in the architecture design.
- Referring to the file list, implement correct import statements.
- Implement all the logic and functions, without any placeholder like "pass", "TODO", etc.
"""

QUALITY_ASSURANCE = """
You are an expert quality assurance engineer from an software development team called "Agile Mind". This team follows the agile model to develop software.

You are given the architecture design and the content of a file that has just been developed. Your job is to review the code quality, find the inconsistencies and potential bugs, and provide feedback to the developers.

Follow these steps:
1. Read and understand the architecture design and the file content carefully.
2. Review the code quality and find all the inconsistencies and potential bugs.
3. Instruct the developers to fix the bugs if any, by providing descriptions and possible solutions. If the code is of high quality, handoff the file to the project manager.

Output the feedback to the developers, or handoff to the project manager.

Note that:
- Focus on obvious bugs and inconsistencies. System testing is going to be done later.
"""

DEBUGGING = """
You are an expert quality assurance engineer from an software development team called "Agile Mind". This team follows the agile model to develop software.

You are given the file path list of a software repository. Your job is to review the code quality of the software repository, find the potential bugs, and instruct the developers to fix them.

Follow these steps:
1. Run static code analysis tools to check the code quality.
2. According to the static analysis results, review the code and find all the potential bugs.
3. Instruct the developers to fix the bugs by providing the file path, bug description, and possible solutions.

Output the bugs in the following JSON format, without any extra characters:
{
    "file_path": "path/to/file",
    "bugs": [
        {
            "line": 10,
            "bug": "potential bug description",
            "solution": "possible solution"
        },
        ...
    ]
}

Note that:
- The static code analysis tools may provide incorrect results or errors from other files. You should review the code carefully to find the potential bugs.
- The bugs should be clear and concise, so that the developers can understand and fix them quickly.
- Other than syntax errors, you should also consider logical errors. Your goal is to make the software work correctly.
"""

DOCUMENT_WRITER = """
You are an expert project manager from an software development team called "Agile Mind". This team follows the agile model to develop software.

Your job is to write a series of documents to summarize the project usage and the key information. You will be given the client's demand and the repository structure.

Follow these steps:
1. Read and understand the demand and the repository structure carefully.
2. Write the following documents:
    - A "README.md" file, located in the root directory, which contains the project introduction, installation guide, usage guide, etc.
    - A "CHANGELOG.md" file, located in the root directory, which contains the current version for future development.

Note that:
- Use the provided tools to write files.
- The README file should be clear and concise, so that the user can understand the project quickly. The target project is local, so you do not need to consider git or CI/CD.
"""

CONTEXT_MAINTAINER = """
You are an expert record keeper from an software development team called "Agile Mind". This team follows the agile model to develop software.

Given the extremely long development process, you are responsible to extract the key information to summarize the project status and progress. 

Note that:
- Do not lose or change any information.
"""


FEEDBACK_TEMPLATE = """
Original demand:
>>>
{raw_demand}
<<<
and the feedback from the UI design:
>>>
{feedback}
<<<
Considering the feedback, the final version of the UI design is:
>>>
{prototype}
<<<
"""

DEVELOPING_TEMPLATE = """
The architecture design is:
>>>
{architecture}
<<<
and your task is to develop the file:
>>>
{file_path}
<<<
"""

QA_FEEDBACK_TEMPLATE = """
Group QA feedback:
In your previously implemented file {file_path}, there are some potential bugs and inconsistencies. Please fix them:
>>>
{description}
<<<
"""



"""
Development of software using waterfall methodology.
"""

import os
import json
import time
import shutil
import readchar
from pathlib import Path
from rich.panel import Panel
from rich.align import Align
from rich.table import Table
from rich.box import ROUNDED
from datetime import timedelta
from typing import Dict, Optional
from rich import print as rich_print
from agilemind.context import Context
from agilemind.tool import get_all_tools
from agilemind.prompt import waterfall_prompt
from agilemind.checker import static_checkers
from agilemind.utils import extract_json, format_cost
from agilemind.execution import Agent, deterministic_generation
from concurrent.futures import ThreadPoolExecutor, as_completed
from rich.progress import Progress, SpinnerColumn, TextColumn, TimeElapsedColumn

debugger = Agent(
    "debugger",
    "debug software",
    waterfall_prompt.PROGRAMMER_DEBUGGING,
    tools=get_all_tools("file_system", "development"),
    generation_params=deterministic_generation,
)
quality_assurance = Agent(
    "quality_assurance",
    "assure software quality",
    waterfall_prompt.QUALITY_ASSURANCE,
    generation_params=deterministic_generation,
)
syntax_debugger = Agent(
    "syntax_debugger",
    "debug syntax",
    waterfall_prompt.SYNTAX_DEBUGGER,
    tools=get_all_tools("file_system", "development"),
    generation_params=deterministic_generation,
)
logic_programmer = Agent(
    "logic_programmer",
    "implement software logic",
    waterfall_prompt.PROGRAMMER_LOGIC,
    tools=get_all_tools("file_system", "development"),
    generation_params=deterministic_generation,
)
structure_programmer = Agent(
    "structure_programmer",
    "implement software structure",
    waterfall_prompt.PROGRAMMER_FRAMEWORK,
    tools=get_all_tools("file_system", "development"),
    generation_params=deterministic_generation,
)
architect = Agent(
    "architect",
    "create software architecture",
    waterfall_prompt.ARCHITECT,
    save_path="docs/software_architecture.json",
    generation_params=deterministic_generation,
)
demand_analyst = Agent(
    "demand_analyst",
    "analyze user demand",
    waterfall_prompt.DEMAND_ANALYST,
    save_path="docs/demand_analysis.md",
    generation_params=deterministic_generation,
)

all_agents = [
    syntax_debugger,
    debugger,
    quality_assurance,
    logic_programmer,
    structure_programmer,
    architect,
    demand_analyst,
]


def run_workflow(
    demand: str,
    max_iterations: int = 5,
    model: Optional[str] = None,
) -> dict:
    """
    Run the LLM-Agent workflow pipelines.

    Args:
        demand: User demand for the software
        max_iterations: Maximum number of iterations for each agent
        model: Model to use for all agents

    Returns:
        Dictionary containing the software development process
    """
    # Set the model for all agents
    if model:
        for agent in all_agents:
            agent.set_model(model)

    output = os.path.abspath(".")
    context = Context(demand, output)
    start_time = time.time()

    with Progress(
        SpinnerColumn(finished_text="[bold green]\N{HEAVY CHECK MARK}"),
        TimeElapsedColumn(),
        TextColumn("[bold blue]{task.description}"),
    ) as progress:
        # Demand analysis step
        demand_task = progress.add_task("Analyzing user demand...", total=1)
        demand_analysis = demand_analyst.process(context, demand, max_iterations)
        progress.update(
            demand_task,
            completed=1,
            description="[bold green]Demand analysis completed",
        )
        context.set_document("demand_analysis", demand_analysis[-1]["output"])
        context.add_history("demand_analysis", demand_analysis)
        demand_report = extract_json(demand_analysis[-1]["output"])

        # Architecture step
        arch_task = progress.add_task("Building architecture...", total=1)
        architecture = architect.process(
            context, json.dumps(demand_analysis), max_iterations
        )
        context.add_history("architecture", architecture)
        architecture = extract_json(architecture[-1]["output"])
        progress.update(
            arch_task, completed=1, description="[bold green]Architecture created"
        )
        context.set_document("architecture", json.dumps(architecture))

        # Implement modules
        modules = architecture.get("modules", [])
        modules_task = progress.add_task("Implementing modules...", total=1)
        inplemented_modules = structure_programmer.process(
            context,
            json.dumps(demand_report) + "\n\n" + json.dumps(modules),
            max_iterations,
        )
        context.add_history("modules", inplemented_modules)
        progress.update(
            modules_task, completed=1, description="[bold green]Modules implemented"
        )

        # Implement the logic of every file in parallel
        files = list(context.code.uptodated.keys())
        logic_task = progress.add_task("Implementing code logic...", total=len(files))
        completed_count = 0
        files_subtasks = []

        def process_code_logic(file: str) -> tuple[str, Dict]:
            """
            Process the logic of a single file.

            Args:
                file: File to process

            Returns:
                Tuple containing the file name and the implemented logic
            """
            file_logic_subtask = progress.add_task(
                f"    Implementing logic for {file}...", total=1
            )
            files_subtasks.append(file_logic_subtask)

            # Prepare the input file data in XML format
            file_data = context.code.uptodated[file]
            xml_data = f"<path>{file}</path>\n<code>{file_data}</code>"

            logic = logic_programmer.process(context, xml_data, max_iterations)
            context.add_history(f"code_logic_{file}", logic)

            progress.update(
                file_logic_subtask,
                completed=1,
                description=f"    [bold green]Logic for {file} implemented",
            )

            return file, logic

        # Execute code logic implementations in parallel
        with ThreadPoolExecutor() as executor:
            future_to_file = {
                executor.submit(process_code_logic, file): file for file in files
            }
            for future in as_completed(future_to_file):
                _, _ = future.result()
                completed_count += 1
                progress.update(
                    logic_task,
                    description=(
                        ("[bold green]" if completed_count == len(files) else "")
                        + f"Logic of {completed_count}/{len(files)} files implemented"
                    ),
                    completed=completed_count,
                )

        for subtask in files_subtasks:
            progress.remove_task(subtask)

        # Check the syntax of the implemented code
        syntax_task = progress.add_task(
            "Checking code syntax...", total=len(context.code.uptodated)
        )
        completed_count = 0
        syntax_subtasks = []

        def check_syntax(file: str) -> tuple[str, dict]:
            """
            Check syntax of a single file.

            Args:
                file: File to check

            Returns:
                Tuple containing the file name and any syntax errors found
            """
            file_syntax_subtask = progress.add_task(
                f"    Checking syntax for {file}...", total=1
            )
            syntax_subtasks.append(file_syntax_subtask)

            error_info = f"in {file}: \n\n"
            static_checkers_result = static_checkers.run(context.code.uptodated[file])

            if static_checkers_result:
                error_info += json.dumps(static_checkers_result)
                context.add_history(f"static_check_{file}", static_checkers_result)
                syntax_debugger.process(context, error_info, max_iterations)

            progress.update(
                file_syntax_subtask,
                completed=1,
                description=f"    [bold green]Syntax for {file} checked",
            )

            return file, static_checkers_result

        # Execute syntax checks in parallel
        with ThreadPoolExecutor() as executor:
            future_to_file = {
                executor.submit(check_syntax, file): file
                for file in context.code.uptodated.keys()
            }
            for future in as_completed(future_to_file):
                _, _ = future.result()
                completed_count += 1
                progress.update(
                    syntax_task,
                    description=(
                        ("[bold green]" if completed_count == len(files) else "")
                        + f"Syntax of {completed_count}/{len(files)} files checked"
                    ),
                    completed=completed_count,
                )

        # Clean up subtasks
        for subtask in syntax_subtasks:
            progress.remove_task(subtask)

        # Quality assurance step
        qa_round = 1
        qa_task = progress.add_task("Assuring software quality...", total=1)
        qa_report = quality_assurance.process(
            context, json.dumps(list(context.code.uptodated.keys())), max_iterations
        )
        context.add_history("quality_assurance", qa_report)
        progress.update(
            qa_task,
            completed=1,
            description=f"[bold green]Quality assurance round {qa_round} completed",
        )
        qa_report = extract_json(qa_report[-1]["output"])

        # Fix the bugs if any
        if qa_report.get("is_buggy", False):
            bugs = qa_report.get("bugs", [])
            debugger_task = progress.add_task("Debugging software...", total=len(bugs))
            debug_result = debugger.process(
                context, json.dumps(qa_report), max_iterations
            )
            progress.remove_task(debugger_task)
            context.add_history("debugger", debug_result)

    # Software information
    total_time = time.time() - start_time
    time_str = str(timedelta(seconds=int(total_time)))
    software_name = architecture["name"]
    module_count = max(len(modules), 1)
    file_count = len(context.code.uptodated.keys())
    lines_of_code = sum(len(f.split("\n")) for f in context.code.uptodated.values())
    doc_count = len(context.document)
    doc_lines = sum(len(doc.split("\n")) for doc in context.document.values())

    # Get token usage info
    token_usage = context.token_usage.to_dict()
    total_prompt_tokens = token_usage["total"]["prompt_tokens"]
    total_completion_tokens = token_usage["total"]["completion_tokens"]
    total_tokens = token_usage["total"]["total_tokens"]

    # Get cost info
    cost_info = context.cost.to_dict()
    total_prompt_cost = cost_info["total"]["prompt_cost"]
    total_completion_cost = cost_info["total"]["completion_cost"]
    total_cost = cost_info["total"]["total_cost"]

    # Create project info table with expanded width
    project_table = Table(
        title="Project Information",
        expand=True,
        padding=(0, 2),
        box=ROUNDED,
        title_justify="left",
    )
    project_table.add_column("Project", style="bold cyan")
    project_table.add_column("Development Time", style="bold yellow")
    project_table.add_column("Project Directory", style="bold green")
    project_table.add_row(software_name, time_str, output)

    # Second row of information
    details_table = Table(
        title="Development Details",
        expand=True,
        padding=(0, 2),
        box=ROUNDED,
        title_justify="left",
        show_header=False,
        show_lines=True,
    )
    details_table.add_row(
        f"Modules Created: [bold yellow]{module_count}[/bold yellow]",
        f"Files Created: [bold blue]{file_count}[/bold blue]",
        f"Documents Created: [bold green]{doc_count}[/bold green]",
    )
    details_table.add_row(
        f"Average files per module: [bold yellow]{file_count / module_count:.2f}[/bold yellow]",
        f"Total Lines of Code: [bold magenta]{lines_of_code}[/bold magenta]",
        f"Total Lines of Document: [bold red]{doc_lines}[/bold red]",
    )

    # Create usage table with expanded width
    usage_table = Table(
        title="Usage Statistics",
        expand=True,
        padding=(0, 2),
        box=ROUNDED,
        title_justify="left",
    )
    usage_table.add_column(style="bold")
    usage_table.add_column("Prompt")
    usage_table.add_column("Completion")
    usage_table.add_column("Total")
    usage_table.add_row(
        "Tokens",
        str(total_prompt_tokens),
        str(total_completion_tokens),
        str(total_tokens),
    )
    usage_table.add_row(
        "Cost",
        "≈ " + format_cost(total_prompt_cost),
        "≈ " + format_cost(total_completion_cost),
        "[bold blue]" + "≈ " + format_cost(total_cost),
    )

    main_table = Table(box=None, expand=True)
    main_table.add_row(project_table)
    main_table.add_row()
    main_table.add_row(details_table)
    main_table.add_row()
    main_table.add_row(usage_table)
    main_table.add_row()

    rich_print(
        Panel(
            Align.center(main_table),
            border_style="bold green",
            title="[bold]\N{HEAVY CHECK MARK} Development Completed Successfully[/bold]",
        )
    )

    return context.dump()


def dev(demand: str, output: str, model: str, max_iterations: int) -> dict:
    """
    Run the LLM-Agent workflow pipelines.

    Args:
        demand: User demand for the software
        output: Directory path to save the software
        model: String name of the model to use
        max_iterations: Maximum number of iterations to run

    Returns:
        Dictionary containing the software development process
    """
    # If output dir exists, ask user whether to confirm purging it first
    if Path(output).exists():
        rich_print(
            Panel(
                Align.center(
                    f'The output directory "{output}" already exists. Do you want to delete its contents? (Y/n)'
                ),
                border_style="bold red",
                title="Warning",
            )
        )

        confirm = readchar.readchar().lower()
        if confirm != "y":
            return {"status": "cancelled"}

        # Purge the output directory
        shutil.rmtree(output)

    Path(output).mkdir(parents=True, exist_ok=True)

    # Change current working directory to the output directory
    initial_cwd = os.getcwd()
    os.chdir(output)

    try:
        result = run_workflow(demand, model=model, max_iterations=max_iterations)

        with open("logs/development_record.json", "w") as f:
            f.write(json.dumps(result, indent=4))
    finally:
        os.chdir(initial_cwd)  # Restore original working directory

    return result



from .syntax_checker import SyntaxChecker
from .import_checker import ImportChecker
from .pylint_checker import PylintChecker
from .checker_pipeline import CheckerPipeline
from .attribute_checker import AttributeChecker

static_checkers = CheckerPipeline()
static_checkers.add_checker(
    SyntaxChecker(), ImportChecker(), AttributeChecker(), PylintChecker()
)



import re
import json
from typing import Optional


def clean_json_string(input_string: str) -> str:
    """
    Clean a string that might contain JSON with markdown or explanatory text.

    Args:
        input_string (str): A string that contains JSON, possibly with markdown formatting or explanatory text around it.

    Returns:
        out (str): A clean JSON string ready to be parsed with json.loads()

    Examples:
        ```
        clean_json_string('json: {"key": "value"}')
        {"key": "value"}
        clean_json_string('Sure! I will... {"key": "value"}')
        {"key": "value"}
        ```
    """
    if not input_string:
        return ""

    json_pattern = r"({[\s\S]*}|\[.*\])"
    json_matches = re.findall(json_pattern, input_string)

    for match in json_matches:
        try:
            json.loads(match)
            return match
        except json.JSONDecodeError:
            continue

    # If fails, return the original string but stripped of common prefixes
    cleaned = re.sub(r"^.*?({.*)", r"\1", input_string, flags=re.DOTALL)
    cleaned = re.sub(r"(}[^}]*)$", r"}", cleaned, flags=re.DOTALL)

    try:
        json.loads(cleaned)
        return cleaned
    except json.JSONDecodeError:
        # If we can't find valid JSON, return the original string
        # as it might be malformed but the caller might want to handle it
        return input_string.strip()


def extract_json(input_string: str) -> Optional[dict]:
    """
    Extract and parse JSON from a string that might have markdown or explanatory text.

    Args:
        input_string (str): A string that contains JSON, possibly with markdown formatting or explanatory text around it.

    Returns:
        out (Optional[dict]): The parsed JSON as a dictionary, or None if parsing fails
    """
    try:
        cleaned = clean_json_string(input_string)
        return json.loads(cleaned)
    except json.JSONDecodeError:
        return {}



"""
Utility functions for loading configuration files and substituting environment variables.
"""

import os
import re
import yaml
from typing import Dict, Any
from dotenv import load_dotenv


def load_config(path: str = "config.yaml") -> Dict[str, Any]:
    """
    Load a YAML configuration file and substitute environment variables.

    Args:
        path (str): Path to the YAML config file. Defaults to "config.yaml" at project root.

    Returns:
        dict: The loaded configuration with environment variables substituted.
    """
    load_dotenv()

    if not os.path.isabs(path):
        project_root = os.path.dirname(
            os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        )
        path = os.path.join(project_root, path)

    # Load the YAML file
    with open(path, "r") as file:
        config = yaml.safe_load(file)

    # Substitute environment variables
    config = _substitute_env_vars(config)

    return config


def _substitute_env_vars(obj: Any) -> Any:
    """
    Recursively substitute environment variables in strings within an object.

    Args:
        obj: The object to process (can be a dict, list, or scalar value)

    Returns:
        The processed object with environment variables substituted
    """
    if isinstance(obj, dict):
        return {k: _substitute_env_vars(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [_substitute_env_vars(item) for item in obj]
    elif isinstance(obj, str):
        # Pattern matches ${VAR_NAME}
        pattern = r"\$\{([^}]+)\}"

        def replace_env_var(match):
            var_name = match.group(1)
            return os.environ.get(var_name, match.group(0))

        return re.sub(pattern, replace_env_var, obj)
    else:
        return obj


def extract_agent_llm_config(agent_name: str, config: dict) -> dict:
    """
    Extract LLM configuration for a specific agent from the config file.

    Args:
        agent_name (str): Name of the agent
        config (dict): Configuration dictionary

    Returns:
        Dictionary containing the LLM configuration for the agent
    """
    config = config or {}

    default_config = config.get("default") or {}
    default_llm_config = default_config.get("llm") or {}
    defalt_base_url = default_llm_config.get("api_base_url")
    default_model = default_llm_config.get("model")

    workflow_config = config.get("workflow") or {}
    agent_config = workflow_config.get(agent_name) or {}
    agent_base_url = agent_config.get("api_base_url") or defalt_base_url
    agent_api_key = agent_config.get("api_key") or None
    agent_model = agent_config.get("model") or default_model

    return {
        "llm_api_key": agent_api_key,
        "llm_base_url": agent_base_url,
        "model": agent_model,
    }



import re
import ast
from .interface import AbsChecker
from typing import Dict, List, Union, Tuple, Optional


def check_syntax(code: str) -> Dict[str, Union[bool, str]]:
    """
    Checks if a given Python code string contains syntax errors by attempting to build an AST.

    Args:
        code (str): The Python code string to check

    Returns:
        out (Dict[str, Union[bool, str]]):
            - 'valid' (bool): True if no syntax errors, False otherwise
            - 'error' (str, optional): Error message if syntax is invalid
    """
    result = {
        "valid": False,
        "error": None,
    }

    try:
        # Attempt to parse the code into an AST
        ast.parse(code)
        result["valid"] = True
        return result
    except SyntaxError as e:
        # Capture syntax error details
        error_line = code.splitlines()[e.lineno - 1]
        result["error"] = (
            f'Syntax error at line {e.lineno} column {e.offset}, "{error_line}": {e.msg}'
        )
        return result
    except Exception as e:
        # Catch any other parsing errors
        result["error"] = f"Error parsing code: {str(e)}"
        return result


def get_syntax_error_location(code: str) -> Optional[Tuple[int, int, str]]:
    """
    Returns the location and message of a syntax error in the provided code.

    Args:
        code (str): The Python code string to check

    Returns:
        out (Optional[Tuple[int, int, str]]): A tuple of (line, column, error_message) if there's a syntax error, None if the code is valid.
    """
    try:
        ast.parse(code)
        return None  # No syntax error
    except SyntaxError as e:
        return (e.lineno, e.offset, e.msg)
    except Exception as e:
        # For non-syntax errors, return position as 1,0
        return (1, 0, str(e))


def has_syntax_error(code: str) -> bool:
    """
    A simple check to determine if the code has any syntax errors.

    Args:
        code (str): The Python code string to check

    Returns:
        out (bool): True if there are syntax errors, False otherwise
    """
    try:
        ast.parse(code)
        return False
    except:
        return True


class SyntaxChecker(AbsChecker):
    """
    A checker that verifies if the given Python code string contains syntax errors.
    """

    @property
    def name(self) -> str:
        return "Syntax Checker"

    def check(self, code: str) -> List[Dict[str, str]]:
        error = check_syntax(code)
        if error["valid"]:
            return []
        pattern = r"Syntax error at line (\d+) column (\d+), \"(.+)\": (.+)"
        return [
            {
                "line": int(match.group(1)),
                "column": int(match.group(2)),
                "problematic_code": match.group(3),
                "message": match.group(4),
            }
            for match in re.finditer(pattern, error["error"])
        ]



"""
Context module for managing state and data flow throughout the pipeline execution.
"""

from .cost import Cost
from datetime import datetime
from .token_usage import TokenUsage
from typing import Any, Dict, List, Optional


class ContextCode:
    """
    Context class that manages the code structure in the context.
    """

    structure: Dict[str, str] = {}
    uptodated: Dict[str, str] = {}

    def dump(self) -> Dict[str, str]:
        """
        Dump the code structure into a dictionary.

        Returns:
            Dictionary containing the code structure
        """
        return {"structure": self.structure, "uptodated": self.uptodated}


class Context:
    """
    Context class that manages the state and data flow throughout pipeline execution.
    """

    root_dir: str
    raw_demand: str
    document: Dict[str, str] = {}
    code: ContextCode = ContextCode()
    history: List[Dict[str, Any]] = []
    time: float = 0.0
    token_usage: TokenUsage
    cost: Cost

    def __init__(self, raw_demand: str, root_dir: Optional[str] = None):
        """
        Initialize the context with the root directory and raw demand.

        Args:
            raw_demand: User demand for the software
            root_dir: Root directory path to save the software
        """
        self.root_dir = root_dir
        self.raw_demand = raw_demand
        self.time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        self.token_usage = TokenUsage()
        self.cost = Cost()

    def is_root_dir_set(self) -> bool:
        """Check if the root directory is set in the context."""
        return self.root_dir is not None

    def set_document(self, key: str, value: str) -> None:
        """
        Set a document in the context.

        Args:
            key: Key to identify the document
            value: Document content

        Returns:
            None
        """
        self.document[key] = value

    def get_document(self, key: str) -> str:
        """
        Get a document from the context.

        Args:
            key: Key to identify the document

        Returns:
            Document content
        """
        return self.document[key]

    def add_history(self, step: str, data: Dict[str, Any]) -> None:
        """
        Add a step to the history in the context.

        Args:
            step: Step name
            data: Data associated with the step

        Returns:
            None
        """
        self.history.append(
            {
                "step": step,
                "time": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                "data": data,
            }
        )

    def update_token_usage(
        self,
        prompt_tokens: int,
        completion_tokens: int,
        agent_name: str,
        round_number: int,
        model: str = None,
    ) -> None:
        """
        Update the token usage in the context with fine-grained tracking.

        Args:
            prompt_tokens: Number of prompt tokens used
            completion_tokens: Number of completion tokens used
            agent_name: Name of the agent making the call
            round_number: Round number within the agent's processing
            model: Optional model name used for the API call
        """
        self.token_usage.update(
            prompt_tokens=prompt_tokens,
            completion_tokens=completion_tokens,
            agent_name=agent_name,
            round_number=round_number,
            model=model,
        )

    def update_cost(
        self,
        prompt_cost: float,
        completion_cost: float,
        agent_name: str,
        round_number: int,
        model: str = None,
    ) -> None:
        """
        Update the cost tracking in the context with fine-grained tracking.

        Args:
            prompt_cost: Cost for prompts
            completion_cost: Cost for completions
            agent_name: Name of the agent making the call
            round_number: Round number within the agent's processing
            model: Optional model name used for the API call
        """
        self.cost.update(
            prompt_cost=prompt_cost,
            completion_cost=completion_cost,
            agent_name=agent_name,
            round_number=round_number,
            model=model,
        )

    def dump(self) -> Dict[str, Any]:
        """
        Dump the context data into a dictionary.

        Returns:
            Dictionary containing the context data
        """
        return {
            "time": self.time,
            "root_dir": self.root_dir,
            "raw_demand": self.raw_demand,
            "document": self.document,
            "code": self.code.dump(),
            "history": self.history,
            "token_usage": self.token_usage.to_dict(),
            "cost": self.cost.to_dict(),
        }



import os
import pathlib
from .definition import Task
from agilemind.context import Context


def save_result_to_file(task: Task, context: Context) -> None:
    """
    Save text content to a file at the given relative path.

    Args:
        file_path (str): The relative path to save the file to.
        content (str): The text content to save.

    Raises:
        ValueError: If the path is not relative or contains ".."
        FileExistsError: If the file already exists.
    """
    file_path = task.artifact_path
    if not file_path:
        raise ValueError("File path not set in task")

    # Check if path is relative and doesn't contain ".."
    if os.path.isabs(file_path):
        raise ValueError("File path must be relative, not absolute")

    if ".." in pathlib.Path(file_path).parts:
        raise ValueError("File path must not contain parent directory references (..)")

    if not context.is_root_dir_set():
        raise ValueError("Root directory not set in context")

    # Get the full path to the file
    file_path = os.path.join(context.root_dir, file_path)

    # Check if file already exists
    if os.path.exists(file_path):
        raise FileExistsError(f"File already exists at {file_path}")

    # Create parent directories if they don't exist
    directory = os.path.dirname(file_path)
    if directory and not os.path.exists(directory):
        os.makedirs(directory)

    # Write content to file
    with open(file_path, "w", encoding="utf-8") as file:
        file.write(task.result.output)



import re
import tempfile
import subprocess
from .interface import AbsChecker
from typing import Any, List, Dict


class PylintError(Exception):
    """Exception raised for errors in the Pylint execution."""

    pass


def run_pylint_on_file(file_path: str) -> str:
    """
    Run pylint on the given file and return the raw output.

    Args:
        file_path (str): The Python file to check

    Returns:
        str: Raw output from pylint

    Raises:
        PylintError: If pylint execution fails
    """
    try:
        pylint = subprocess.run(
            ["which", "pylint"], text=True, capture_output=True, check=True
        )
        pylint = pylint.stdout.strip()
        result = subprocess.run(
            [pylint, file_path, "-E", "--disable=E0401"],
            text=True,
            capture_output=True,
        )
        return result.stdout + result.stderr
    except Exception as e:
        raise PylintError(f"Failed to execute pylint: {str(e)}")


def run_pylint_on_code(code_str: str) -> str:
    """
    Run pylint on the given code string and return the raw output.

    Args:
        code_str (str): The Python code to check

    Returns:
        str: Raw output from pylint

    Raises:
        PylintError: If pylint execution fails
    """
    # Save code to a temporary file
    with tempfile.NamedTemporaryFile(suffix=".py") as temp_file:
        temp_file_path = temp_file.name
        temp_file.write(code_str.encode("utf-8"))
        temp_file.flush()

        return run_pylint_on_file(temp_file_path)


def parse_pylint_errors(pylint_output: str) -> List[Dict[str, Any]]:
    """
    Parse pylint output and extract error information.

    Args:
        pylint_output (str): Raw output from pylint

    Returns:
        out (List[Dict[str, Any]]): List of error dictionaries
    """
    # Regular expression to extract error information
    # Format typically: filename:line:column: error_code: message
    error_pattern = r"(.*?):(\d+):(\d+): (\w+): (.+)"

    errors = []
    for line in pylint_output.split("\n"):
        if not line.strip():
            continue

        match = re.search(error_pattern, line)
        if match and match.group(4).startswith("E"):  # Ensure it's an error (E)
            errors.append(
                {
                    "file": match.group(1),
                    "line": int(match.group(2)),
                    "column": int(match.group(3)),
                    "code": match.group(4),
                    "message": match.group(5),
                }
            )

    return errors


def check_code_with_pylint(code_str: str) -> List[Dict[str, Any]]:
    """
    Check the given Python code string with PyLint and return errors (Exxxx).

    Args:
        code_str (str): A string representation of Python code to be checked.

    Returns:
        out (List[Dict[str, Any]]): A list of dictionaries containing error details:
            - code: The error code (e.g., E0001)
            - line: Line number where the error occurred
            - column: Column number where the error occurred
            - message: Description of the error
    """
    try:
        pylint_output = run_pylint_on_code(code_str)
        return parse_pylint_errors(pylint_output)
    except PylintError as e:
        return [
            {
                "file": "",
                "code": "Unknown",
                "line": 0,
                "column": 0,
                "message": f"Error running PyLint: {str(e)}",
            }
        ]


def check_file_with_pylint(file_path: str) -> List[Dict[str, Any]]:
    """
    Check the given Python file with PyLint and return errors (Exxxx).

    Args:
        file_path (str): A string representation of Python file to be checked.

    Returns:
        out (List[Dict[str, Any]]): A list of dictionaries containing error details:
            - code: The error code (e.g., E0001)
            - line: Line number where the error occurred
            - column: Column number where the error occurred
            - message: Description of the error
    """
    try:
        pylint_output = run_pylint_on_file(file_path)
        return parse_pylint_errors(pylint_output)
    except PylintError as e:
        return [
            {
                "file": file_path,
                "code": "Unknown",
                "line": 0,
                "column": 0,
                "message": f"Error running PyLint: {str(e)}",
            }
        ]


class PylintChecker(AbsChecker):
    """
    A checker that verifies Python code using PyLint.
    """

    @property
    def name(self) -> str:
        return "PyLint Checker"

    def check(self, code: str) -> List[Dict[str, str]]:
        errors = check_code_with_pylint(code)
        return [
            {
                "line": error["line"],
                "column": error["column"],
                "problematic_code": code.splitlines()[error["line"] - 1],
                "message": error["message"],
            }
            for error in errors
        ]



"""
Agent module for creating agents that can process inputs, use tools, and hand off to other agents.
"""

import os
import json
import openai
from dotenv import load_dotenv
from .config import GenerationParams
from agilemind.context import Context
from agilemind.tool import execute_tool
from typing import List, Optional, Dict, Union
from agilemind.utils import retry, calculate_cost, clean_json_string


load_dotenv()


class Agent:
    """
    An agent that can perform tasks based on instructions and use tools or hand off to other agents.

    Agents can process inputs according to their instructions, use tools to perform
    actions, or hand off to other specialized agents when appropriate.
    """

    def __init__(
        self,
        name: str,
        description: str,
        instructions: str,
        tools: Optional[List[Dict[str, str]]] = None,
        handoffs: Optional[List["Agent"]] = None,
        next_agent: Optional["Agent"] = None,  # Added next_agent for forced handoff
        model: str = "gpt-4o-mini",
        save_path: Optional[str] = None,  # Path to save agent responses
        generation_params: Optional[Union[GenerationParams, Dict]] = None,
        llm_base_url: Optional[str] = None,
        llm_api_key: Optional[str] = None,
    ):
        """
        Initialize an Agent instance.

        Args:
            name: The name of the agent. Should be unique, lowercase, and without spaces.
            description: Brief description of what the agent does
            instructions: Instructions that define the agent's behavior.
            tools: Optional list of tools the agent can use.
            handoffs: Optional list of agents this agent can hand off to.
            next_agent: Optional next agent for forced handoff regardless of the agent's decision.
            model: OpenAI model to use for this agent
            save_path: Optional path to save agent's responses for documentation purposes
            generation_params: Optional parameters for the model generation
            llm_base_url: Optional base URL for the OpenAI API
            llm_api_key: Optional API key for the OpenAI API
        """
        self.name = name
        self.description = description
        self.instructions = instructions
        self.tools = tools or []
        self.handoffs = handoffs or []
        self.next_agent = next_agent  # Store the next agent for forced handoff
        self.model = model
        self.save_path = save_path
        self.rounds = []  # Track information by round
        self.generation_params = generation_params

        if isinstance(generation_params, dict):
            self.generation_params = GenerationParams(**generation_params)

        self.client = openai.OpenAI(
            api_key=llm_api_key or os.getenv("OPENAI_API_KEY"),
            base_url=llm_base_url or os.getenv("OPENAI_API_BASE_URL"),
        )

    def __repr__(self) -> str:
        """Return string representation of the Agent."""
        return f"Agent(name='{self.name}')"

    def get_available_tools(self) -> List[Dict[str, str]]:
        """Return the list of tools available to this agent."""
        return self.tools

    def get_available_handoffs(self) -> List["Agent"]:
        """Return the list of agents this agent can hand off to."""
        return self.handoffs

    def set_model(self, model: str) -> None:
        """
        Set the model to use for this agent.

        Args:
            model (str): The model to use
        """
        self.model = model

    def save_response(self, response_content: str) -> None:
        """
        Save the agent's response to the specified path.

        Args:
            response_content: The content to save to the file
        """
        if not self.save_path:
            return

        # Create directory if it doesn't exist
        os.makedirs(os.path.dirname(self.save_path), exist_ok=True)

        # Append to file if it exists, otherwise create it
        with open(self.save_path, "a") as f:
            f.write(response_content + "\n\n")

    def process(
        self, context: Context, input_text: str, max_iterations: int = None
    ) -> List[Dict]:
        """
        Process the input using OpenAI API and return the agent's response.
        This method is decorated with retry to handle transient failures.

        Args:
            context: The context object
            input_text: The text input to process
            max_iterations: Maximum number of interaction rounds (None means no limit)

        Returns:
            Dict containing the agent's response and any actions taken
        """
        result = self._process_with_retry(context, input_text, max_iterations)
        context.add_history(self.name, result.copy())
        return result

    @retry(
        exceptions=[
            openai.APIError,
            openai.APIConnectionError,
            openai.RateLimitError,
            openai.APITimeoutError,
            openai.InternalServerError,
            json.JSONDecodeError,
        ],
    )
    def _process_with_retry(
        self, context: Context, input_text: str, max_iterations: int = None
    ) -> List[Dict]:
        """
        Internal method to process input with retry capabilities.
        This method is decorated with retry to handle transient failures.

        Args:
            context: The context object
            input_text: The text input to process
            max_iterations: Maximum number of interaction rounds (None means no limit)

        Returns:
            List of dictionaries containing information for each round of interaction
        """
        messages = self._prepare_messages(input_text)

        # Initialize rounds tracking
        self.rounds = []
        current_round = {
            "input": input_text,
            "output": None,
            "tool_calls": None,
            "handoff": None,
            "token_usage": None,
            "cost": None,  # Add cost field to track costs per round
        }

        # Add handoff agents as tools
        tools_with_handoffs = (
            self.tools.copy()
        )  # Create a copy to avoid modifying the original
        for agent in self.handoffs:
            tools_with_handoffs.append(
                {
                    "type": "function",
                    "function": {
                        "name": f"handoff_to_{agent.name}",
                        "description": f"Hand off to the {agent.name}, who is specialized in {agent.description}.",
                        "parameters": {
                            "type": "object",
                            "properties": {
                                "instructions": {
                                    "type": "string",
                                    "description": "Instructions for that character to follow.",
                                }
                            },
                        },
                    },
                }
            )

        round_number = 0

        # Continue conversation until no more tool calls or a handoff is requested
        # or until max_iterations is reached (if specified)
        while max_iterations is None or round_number < max_iterations:
            round_number += 1

            try:
                print(f"Debug: {self.name} processing round {round_number}...")
                response = self.client.chat.completions.create(
                    model=self.model,
                    messages=messages,
                    tools=tools_with_handoffs if tools_with_handoffs else None,
                    **(
                        self.generation_params.to_dict()
                        if self.generation_params
                        else {}
                    ),
                )
                print("Debug: Response received.")
            except Exception as e:
                print(f"Error: {e}")
                raise

            response_message = response.choices[0].message

            # Update current round with output
            current_round["output"] = response_message.content

            # Record token usage with enhanced details
            if hasattr(response, "usage") and response.usage:
                prompt_tokens = response.usage.prompt_tokens
                completion_tokens = response.usage.completion_tokens

                # Record token usage
                current_round["token_usage"] = {
                    "prompt_tokens": prompt_tokens,
                    "completion_tokens": completion_tokens,
                    "total_tokens": response.usage.total_tokens,
                }

                # Calculate cost using the model_pricing utility
                cost_info = calculate_cost(
                    model=self.model,
                    prompt_tokens=prompt_tokens,
                    completion_tokens=completion_tokens,
                )

                # Record cost information
                current_round["cost"] = cost_info

                # Update token usage in context with detailed information
                context.update_token_usage(
                    prompt_tokens=prompt_tokens,
                    completion_tokens=completion_tokens,
                    agent_name=self.name,
                    round_number=round_number,
                    model=self.model,
                )

                # Update cost information in context
                context.update_cost(
                    prompt_cost=cost_info["prompt_cost"],
                    completion_cost=cost_info["completion_cost"],
                    agent_name=self.name,
                    round_number=round_number,
                    model=self.model,
                )

            # Save the response if a save path is specified
            if self.save_path and response_message.content:
                if self.save_path.endswith(".json"):
                    self.save_response(clean_json_string(response_message.content))
                else:
                    self.save_response(response_message.content)

            # Check for handoff or tool calls
            handoff_requested = False
            current_round_tool_calls = []

            if hasattr(response_message, "tool_calls") and response_message.tool_calls:
                for tool_call in response_message.tool_calls:
                    tool_name = tool_call.function.name

                    # Check if this is a handoff
                    if tool_name.startswith("handoff_to_"):
                        target_agent_name = tool_name[len("handoff_to_") :]
                        args = json.loads(tool_call.function.arguments)
                        instruction = args.get("instructions", "")
                        # TODO: Add support for handoff instructions

                        for agent in self.handoffs:
                            if agent.name == target_agent_name:
                                current_round["handoff"] = agent.name
                                handoff_requested = True
                                break
                        if handoff_requested:
                            break
                    else:
                        # Execute the tool
                        args = json.loads(tool_call.function.arguments)
                        tool_result = execute_tool(context, tool_name, args)
                        current_tool_call = {
                            "tool": tool_name,
                            "args": args,
                            "result": tool_result,
                        }
                        current_round_tool_calls.append(current_tool_call)

                        # Add the tool call and result to messages for the next turn
                        messages.append(
                            {
                                "role": "assistant",
                                "content": response_message.content,
                                "tool_calls": [
                                    {
                                        "id": tool_call.id,
                                        "type": "function",
                                        "function": {
                                            "name": tool_name,
                                            "arguments": tool_call.function.arguments,
                                        },
                                    }
                                ],
                            }
                        )
                        messages.append(
                            {
                                "role": "tool",
                                "tool_call_id": tool_call.id,
                                "content": json.dumps(tool_result),
                            }
                        )

            # Update current round with tool calls if any
            if current_round_tool_calls:
                current_round["tool_calls"] = current_round_tool_calls

            # Add the current round to rounds list
            self.rounds.append(current_round.copy())

            # Break the loop if no tool calls or handoff is requested
            # or if we've reached the maximum number of iterations
            if (
                not current_round_tool_calls
                or handoff_requested
                or (max_iterations is not None and round_number >= max_iterations)
            ):
                break

            # Prepare for next round - tool results become the new input
            next_input = f"Tool results: {json.dumps(tool_result)}"
            current_round = {
                "input": next_input,
                "output": None,
                "tool_calls": None,
                "handoff": None,
                "token_usage": None,
            }

        # Check if there's a forced handoff via next_agent, which takes precedence over
        # any handoff the agent might have selected
        if self.next_agent:
            # Update the last round with the forced handoff
            self.rounds[-1]["handoff"] = self.next_agent.name

        return self.rounds

    def _prepare_messages(self, input_text: str) -> List[Dict]:
        """Prepare the message for the API call."""
        return [
            {"role": "system", "content": self.instructions},
            {"role": "user", "content": input_text},
        ]



"""
This module contains the Executor class that represents an agent executing tasks in a chat session.
It provides methods to interact with the OpenAI chat API and handle query execution.
"""

import json
from openai import OpenAI
from typing import Optional
from agilemind.task import Task
from dataclasses import dataclass
from .config import ExecutorConfig
from agilemind.context import Context
from agilemind.prompt import DEFAULT_SYSTEM_MESSAGE
from agilemind.tool import get_all_tools, execute_tool
from openai.types.chat.chat_completion import ChatCompletion, ChatCompletionMessage


@dataclass
class ExecutionUsage:
    """
    Dataclass to represent the usage statistics of an execution response.
    """

    prompt: int
    completion: int


class Executor:
    """
    Class to represent an executor agent that processes queries through an LLM.
    """

    config: ExecutorConfig
    client: OpenAI

    def __init__(self, config: Optional[ExecutorConfig] = None):
        """
        Initialize the Executor class with a prompt and configuration.

        Parameters:
            prompt (str): The initial prompt for the executor.
            config (ExecutorConfig, optional): Configuration for the executor.
            model (str, optional): The model to use. If provided, overrides the model from config.

        Raises:
            ValueError: If the API key is not provided or the model is not specified.
        """
        # Initialize with config if provided, otherwise load from env
        self.config = config if config else ExecutorConfig.from_env()
        if not self.config.api_key:
            raise ValueError("API key is required")

        # Initialize OpenAI client
        self.client = OpenAI(api_key=self.config.api_key, base_url=self.config.base_url)

    @classmethod
    def from_env(cls) -> "Executor":
        """Create an executor from environment variables."""
        return cls(config=ExecutorConfig.from_env())

    def execute(self, task: Task, context: Context) -> Context:
        """
        Executes a task.

        Args:
            task: The Task instance containing the task description and metadata
            context: Contextual information

        Returns:
           Dict[str, Any]: The response from the model containing the generated text and usage statistics.

        Raises:
            ValueError: If an unsupported generation parameter is provided.
        """
        system_message = task.agent.background or DEFAULT_SYSTEM_MESSAGE
        model = task.agent.model or self.config.default_model
        query = task.input
        task.status = "in_progress"

        # Merge default generation params from config with override params
        merged_params = self.config.generation_params.to_dict()
        merged_params.update(task.agent.config or {})
        if task.agent.use_tool:
            merged_params["tools"] = get_all_tools()

        try:
            response: ChatCompletion = self.client.chat.completions.create(
                messages=[
                    {"role": "system", "content": system_message},
                    {"role": "user", "content": query},
                ],
                model=model,
                **merged_params,
            )
            message: ChatCompletionMessage = response.choices[0].message

            if (
                message.tool_calls
                and len(message.tool_calls) > 0
                and message.tool_calls[0]
            ):
                try:
                    # Handle tool calls
                    tool_call = message.tool_calls[0]
                    func = tool_call.function

                    # Execute the tool function
                    tool_result = execute_tool(func.name, json.loads(func.arguments))
                except:
                    raise ValueError("Invalid tool call format")

            # Assemble the result from the response
            result = {
                "response": response.choices[0].message.content,
                "usage": ExecutionUsage(
                    prompt=response.usage.prompt_tokens,
                    completion=response.usage.completion_tokens,
                ),
            }

            task.set_result(result)

            # Update context with task result
            context[f"task_{task.get_name()}_result"] = result
            return context

        except Exception as e:
            task.set_failed(str(e))
            raise



from typing import List, Dict
from abc import ABC, abstractmethod


class AbsChecker(ABC):
    """
    Abstract base class for code checkers.

    All code checkers should implement this interface.
    """

    @property
    @abstractmethod
    def name(self) -> str:
        """
        The name of the checker.
        """
        pass

    @abstractmethod
    def check(self, code: str) -> List[Dict[str, str]]:
        """
        Check the given code and return a list of issues found.

        Args:
            code (str): The code to check

        Returns:
            out (List[Dict[str, Any]]): A list of dictionaries containing information about issues found.
                Each dictionary should typically contain information like:
                - 'line': line number of the issue
                - 'column': column number of the issue
                - 'problematic_code': code snippet that caused the issue
                - 'message': description of the issue
        """
        pass



def format_cost(cost: float) -> str:
    """
    Format cost to keep only the first non-zero digit after the decimal point.

    Examples:
        ```
        0.00000123 -> 0.000001
        0.123456 -> 0.1
        10.321 -> 10.3
        ```

    Args:
        cost (float): The cost value to format

    Returns:
        Formatted cost as a string
    """
    if cost == 0:
        return "0"

    # Convert to string in scientific notation
    sci_notation = f"{cost:.10e}"
    mantissa, exponent = sci_notation.split("e")
    mantissa = float(mantissa)
    exponent = int(exponent)

    # Find the first significant digit after decimal
    if exponent < 0:
        # For numbers less than 1
        precision = abs(exponent)
        return f"{cost:.{precision}f}"
    else:
        # For numbers >= 1
        # Keep only first digit after decimal
        return f"{cost:.1f}"



from .agent import Agent
from .runner import Runner
from .config import GenerationParams


creative_generation = GenerationParams(
    top_p=0.9,
    temperature=0.8,
)
deterministic_generation = GenerationParams(
    temperature=0.2,
)
neutral_generation = GenerationParams(
    top_p=0.5,
    temperature=0.5,
)

__all__ = [
    "Agent",
    "Runner",
    "GenerationParams",
    "creative_generation",
    "deterministic_generation",
    "neutral_generation",
]



from . import fixed as fixed_prompt
from . import agile as agile_prompt
from . import waterfall as waterfall_prompt

DEFAULT_SYSTEM_MESSAGE = """You are a helpful assistant."""

__all__ = ["DEFAULT_SYSTEM_MESSAGE", "fixed_prompt", "agile_prompt", "waterfall_prompt"]



"""
CLI window display using rich to show live updates.
"""

import uuid
import time
import threading
from rich import box
from rich.live import Live
from rich.text import Text
from rich.rule import Rule
from rich.tree import Tree
from rich.table import Table
from rich.panel import Panel
from datetime import datetime
from rich.console import Group
from rich.console import Console
from typing import Any, Dict, Optional, Literal, List, Tuple, Callable

SPINNER = ["|", "/", "-", "\\"]


def intercept_print(log_window: "LogWindow", level: str = "INFO") -> Callable:
    """
    Intercept the built-in print function and redirect output to the LogWindow.

    Args:
        log_window: The LogWindow instance to send print output to
        level: Log level to use for the intercepted print messages

    Returns:
        A function to restore the original print function
    """
    import builtins

    original_print = builtins.print

    def custom_print(*args, **kwargs):
        # Extract file and sep kwargs if provided
        file = kwargs.get("file", None)
        sep = kwargs.get("sep", " ")

        # If printing to a specific file (not stdout), use original print
        if file is not None:
            original_print(*args, **kwargs)
            return

        # Convert args to string with the specified separator
        message = sep.join(str(arg) for arg in args)

        log_level = level
        if "debug" in message.lower():
            log_level = "DEBUG"
        elif "warning" in message.lower():
            log_level = "WARNING"
        elif "error" in message.lower():
            log_level = "ERROR"
        elif "critical" in message.lower():
            log_level = "CRITICAL"

        # Log the message
        log_window.log(message, level=log_level)

    # Replace the built-in print with our custom function
    builtins.print = custom_print

    # Return a function to restore the original print
    def restore_print():
        builtins.print = original_print
        return original_print

    return restore_print


class LogWindow:
    """Live updating CLI window to display progress and logs."""

    def __init__(
        self,
        title: str = "AgileMind",
        refresh_per_second: float = 5,
        display_style: Literal["tree", "table"] = "tree",
        log_height: int = 5,
        window_height: Optional[int] = None,
        intercept_print: bool = True,
        print_log_level: str = "INFO",
        auto_refresh: bool = True,
    ):
        """
        Initialize the LogWindow.

        Args:
            title (str): The title of the window
            refresh_per_second (float): How many times per second the display refreshes
            display_style (Literal["tree", "table"]): Display style for tasks - "tree" or "table"
            log_height (int): Number of log lines to show in the log zone
            window_height (Optional[int]): Total height of the window. If None, uses full terminal height
            intercept_print (bool): Whether to automatically intercept print function calls
            print_log_level (str): Log level to use for intercepted print messages
            auto_refresh (bool): Whether to automatically refresh the window at regular intervals
        """
        self.title = title
        self.console = Console()
        self.tasks: Dict[str, Dict[str, Any]] = {}
        self.task_hierarchy: Dict[str, Optional[str]] = {}  # task_id -> parent_id
        self._live: Optional[Live] = None
        self.refresh_per_second = refresh_per_second
        self.display_style = display_style
        self.hidden = False
        # Log storage
        self.logs: List[Tuple[datetime, str, str]] = []
        self.log_height = log_height
        self.window_height = window_height
        self.full_screen = window_height is None

        # Auto-refresh settings
        self.auto_refresh = auto_refresh
        self.auto_refresh_interval = 1.0 / refresh_per_second
        self._auto_refresh_thread = None
        self._stop_refresh_thread = threading.Event()

        # Store the print restore function
        self._restore_print = None

        # Intercept print if requested
        if intercept_print:
            self._restore_print = self.intercept_print(level=print_log_level)

        self._calculate_heights()

    def _calculate_heights(self):
        """Calculate the heights of the different zones."""
        if self.full_screen:
            _, height = self.console.size
            self.window_height = height
            self.task_zone_height = self.window_height - self.log_height - 4

    def __enter__(self):
        """Context manager entry point."""
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Context manager exit point."""
        self.close()

    def open(self):
        """Open and display the log window."""
        self._live = Live(
            self._generate_display(),
            console=self.console,
            refresh_per_second=self.refresh_per_second,
            screen=True,
        )
        self._live.start()
        self.hidden = False

        # Start auto-refresh thread if enabled
        if self.auto_refresh:
            self._start_auto_refresh()

        return self

    def close(self):
        """Close the log window."""
        # Stop the auto-refresh thread if it's running
        self._stop_auto_refresh()

        if self._live:
            self._live.stop()
            self._live = None
            self.hidden = False

        # Restore original print function if it was intercepted
        if self._restore_print is not None:
            self._restore_print()
            self._restore_print = None

    def hide(self):
        """Temporarily hide the log window without closing it."""
        # Stop auto-refresh while hidden
        self._stop_auto_refresh()

        if self._live and not self.hidden:
            # Save current state before stopping
            self._saved_display = self._generate_display()
            self._live.stop()
            self._live = None
            self.hidden = True
            # Clear the screen to remove the display
            self.console.clear()

    def show(self):
        """Show the log window if it was hidden."""
        if self.hidden:
            # Restart with the saved display
            self._live = Live(
                (
                    self._saved_display
                    if hasattr(self, "_saved_display")
                    else self._generate_display()
                ),
                console=self.console,
                refresh_per_second=self.refresh_per_second,
                screen=True,
            )
            self._live.start()
            self.hidden = False

            # Restart auto-refresh if enabled
            if self.auto_refresh:
                self._start_auto_refresh()

    def toggle_visibility(self):
        """Toggle the visibility of the log window."""
        if self.hidden:
            self.show()
        else:
            self.hide()

    def add_task(
        self, description: str, parent_id: Optional[str] = None, status: str = "pending"
    ) -> str:
        """
        Add a new task to the window.

        Args:
            description (str): Description of the task
            parent_id (Optional[str]): Optional ID of the parent task
            status (str): Initial status of the task

        Returns:
            task_id: Identifier for the added task
        """
        task_id = str(uuid.uuid4())
        self.tasks[task_id] = {
            "description": description,
            "status": status,
            "time_added": datetime.now(),
        }
        self.task_hierarchy[task_id] = parent_id

        if self._live:
            self._live.update(self._generate_display())

        return task_id

    def complete_task(self, task_id: str):
        """
        Mark a task as completed.

        Args:
            task_id (str): ID of the task to complete
        """
        if task_id in self.tasks:
            self.tasks[task_id]["status"] = "completed"
            self.tasks[task_id]["time_completed"] = datetime.now()

            if self._live:
                self._live.update(self._generate_display())

    def update_task(self, task_id: str, status: str, description: Optional[str] = None):
        """
        Update a task's status and optionally its description.

        Args:
            task_id (str): ID of the task to update
            status (str): New status, e.g., "pending", "running", "completed", "failed"
            description (Optional[str]): New description (if provided)
        """
        if task_id in self.tasks:
            self.tasks[task_id]["status"] = status
            if description:
                self.tasks[task_id]["description"] = description

            if self._live:
                self._live.update(self._generate_display())

    def set_display_style(self, style: Literal["tree", "table"]):
        """
        Set the display style for the tasks.

        Args:
            style (Literal["tree", "table"]): Either "tree" or "table"
        """
        if style not in ["tree", "table"]:
            raise ValueError("Display style must be either 'tree' or 'table'")

        self.display_style = style
        if self._live:
            self._live.update(self._generate_display())

    def log(self, message: str, level: str = "INFO"):
        """
        Add a log message to the log zone.

        Args:
            message (str): The log message to add
            level (str): Log level (INFO, WARNING, ERROR, DEBUG, etc.)
        """
        timestamp = datetime.now()

        # Split multi-line messages and add each line as a separate log
        for line in message.split("\n"):
            if line.strip():
                self.logs.append((timestamp, level, line))

        # Maintain fixed size by removing oldest logs
        if len(self.logs) > self.log_height:
            self.logs = self.logs[-self.log_height :]

        # Update the display if live
        if self._live:
            self._live.update(self._generate_display())

    def clear_logs(self):
        """Clear all logs from the log zone."""
        self.logs = []
        if self._live:
            self._live.update(self._generate_display())

    def intercept_print(self, level: str = "INFO") -> Callable:
        """
        Intercept the built-in print function and redirect output to this LogWindow.

        Args:
            level: Log level to use for the intercepted print messages

        Returns:
            A function to restore the original print function
        """
        return intercept_print(self, level)

    def _start_auto_refresh(self):
        """Start a background thread for auto-refreshing the display."""
        if (
            self._auto_refresh_thread is not None
            and self._auto_refresh_thread.is_alive()
        ):
            return  # Thread already running

        self._stop_refresh_thread.clear()
        self._auto_refresh_thread = threading.Thread(
            target=self._auto_refresh_loop, daemon=True
        )
        self._auto_refresh_thread.start()

    def _stop_auto_refresh(self):
        """Stop the auto-refresh background thread."""
        if (
            self._auto_refresh_thread is not None
            and self._auto_refresh_thread.is_alive()
        ):
            self._stop_refresh_thread.set()
            self._auto_refresh_thread.join(timeout=1.0)
            self._auto_refresh_thread = None

    def _auto_refresh_loop(self):
        """Background thread loop to periodically update the display."""
        while not self._stop_refresh_thread.is_set():
            if self._live and not self.hidden:
                try:
                    self._live.update(self._generate_display())
                except Exception:
                    # Prevent crashes from display updates
                    pass
            time.sleep(self.auto_refresh_interval)

    def set_auto_refresh(self, enabled: bool, interval: Optional[float] = None):
        """
        Enable or disable automatic display refreshing.

        Args:
            enabled (bool): Whether to enable auto-refresh
            interval (Optional[float]): Refresh interval in seconds. If None, uses existing interval.
        """
        self.auto_refresh = enabled
        if interval is not None:
            self.auto_refresh_interval = interval

        if enabled:
            self._start_auto_refresh()
        else:
            self._stop_auto_refresh()

    def _generate_display(self) -> Panel:
        """Generate the display content."""
        self._calculate_heights()

        if self.display_style == "tree":
            max_items = max(1, self.task_zone_height - 3)
            tasks_display = self._generate_task_tree(max_items=max_items)
            tasks_size = len(tasks_display.children) + 3
            tasks_display = Group(Text(""), tasks_display, Text(""))
        else:
            max_rows = max(1, self.task_zone_height - 4)
            tasks_display = self._generate_task_table(max_rows=max_rows)
            tasks_size = tasks_display.row_count + 4

        # If task size is smaller than the window height, pad with empty lines
        missing_lines = self.task_zone_height - tasks_size
        tasks_display = (
            Group(tasks_display, Text("\n" * missing_lines, style="dim"))
            if missing_lines > 0
            else tasks_display
        )

        # Create the log zone
        log_display = self._generate_log_zone()

        # Combine displays using Group which can handle any renderable
        combined_display = Group(tasks_display, Rule(style="dim"), log_display)

        return Panel(
            combined_display,
            title=f"[bold blue]{self.title}[/bold blue]",
            border_style="blue",
            box=box.ROUNDED,
            height=self.window_height,
        )

    def _generate_log_zone(self) -> Text:
        """Generate the log zone display."""
        log_text = Text()

        # If no logs, add placeholder
        if not self.logs:
            log_text.append("No logs", style="dim italic")
            return log_text

        # Add each log line with timestamp and level
        for i, (timestamp, level, message) in enumerate(self.logs):
            if i > 0:
                log_text.append("\n")

            # Format timestamp
            time_str = timestamp.strftime("%H:%M:%S")
            log_text.append(time_str, style="bright_black")

            # Add level with appropriate color
            log_text.append(" - ")
            log_text.append(level, style=self._get_level_style(level))

            # Add message
            log_text.append(" - ")
            log_text.append(message)

        # Pad with empty lines to maintain fixed height
        missing_lines = self.log_height - len(self.logs)
        for _ in range(missing_lines):
            log_text.append("\n")
            log_text.append("", style="dim")

        return log_text

    def _generate_task_tree(self, max_items: Optional[int] = None) -> Tree:
        """Generate a hierarchical tree of tasks.

        Args:
            max_items: Maximum number of items to show in the tree
        """
        tree = Tree("[bold]Task Hierarchy[/bold]")
        count = 0

        # Check if we even need to consider limits
        limited = max_items is not None and len(self.tasks) > max_items

        roots_found = False
        # Process root tasks first
        root_tasks = [
            (task_id, self.tasks[task_id])
            for task_id, parent in self.task_hierarchy.items()
            if parent is None
        ]

        # If we have root tasks, add them
        for task_id, task in root_tasks:
            # Check if we've reached the limit
            if limited and count >= max_items - 1:  # Save space for ellipsis
                tree.add("[dim]...[/dim]")
                count += 1
                break

            roots_found = True
            status_style = self._get_status_style(task["status"])

            # Format task info
            task_text = f"{task['description']} {status_style}"
            time_info = f"[dim](Added: {task['time_added'].strftime('%H:%M:%S')}"
            if "time_completed" in task:
                time_info += (
                    f", Completed: {task['time_completed'].strftime('%H:%M:%S')}"
                )
            time_info += ")[/dim]"

            branch = tree.add(f"{task_text} {time_info}")
            count += 1

            # Add children with remaining space
            remaining = None if not limited else max(0, max_items - count)
            child_count = self._add_child_tasks(branch, task_id, remaining)
            count += child_count

            # If we've reached the limit after adding children, stop
            if limited and count >= max_items:
                break

        # If no root tasks were found, create a flat tree
        if not roots_found and self.tasks:
            flat_task_items = list(self.tasks.items())

            # Calculate how many tasks we can show
            tasks_to_show = (
                len(flat_task_items)
                if not limited
                else min(max_items, len(flat_task_items))
            )

            for task_id, task in flat_task_items[:tasks_to_show]:
                status_style = self._get_status_style(task["status"])
                task_text = f"{task['description']} {status_style}"
                time_info = f"[dim](Added: {task['time_added'].strftime('%H:%M:%S')}"
                if "time_completed" in task:
                    time_info += (
                        f", Completed: {task['time_completed'].strftime('%H:%M:%S')}"
                    )
                time_info += ")[/dim]"

                tree.add(f"{task_text} {time_info}")
                count += 1

            # Add ellipsis if we limited
            if limited and len(flat_task_items) > tasks_to_show:
                tree.add("[dim]...[/dim]")
                count += 1

        return tree

    def _add_child_tasks(
        self, parent_branch, parent_id, remaining: Optional[int] = None
    ) -> int:
        """Add child tasks to a parent branch.

        Args:
            parent_branch: The branch to add children to
            parent_id: ID of the parent task
            remaining: Number of remaining slots available

        Returns:
            int: Number of nodes added
        """
        count = 0

        # Get all child tasks for this parent
        child_tasks = [
            (task_id, self.tasks[task_id])
            for task_id, parent in self.task_hierarchy.items()
            if parent == parent_id
        ]

        # Check for limiting
        limited = remaining is not None and len(child_tasks) > remaining

        # Determine how many children we can show
        tasks_to_show = len(child_tasks) if not limited else remaining

        for i, (task_id, task) in enumerate(child_tasks[:tasks_to_show]):
            status_style = self._get_status_style(task["status"])

            # Format task info
            task_text = f"{task['description']} {status_style}"
            time_info = f"[dim](Added: {task['time_added'].strftime('%H:%M:%S')}"
            if "time_completed" in task:
                time_info += (
                    f", Completed: {task['time_completed'].strftime('%H:%M:%S')}"
                )
            time_info += ")[/dim]"

            branch = parent_branch.add(f"{task_text} {time_info}")
            count += 1

            # Calculate remaining slots for next level
            next_remaining = None if remaining is None else max(0, remaining - count)

            # Add grandchildren if we have space left
            if next_remaining is None or next_remaining > 0:
                grandchild_count = self._add_child_tasks(
                    branch, task_id, next_remaining
                )
                count += grandchild_count

                # If we've used up all slots, stop adding more children
                if remaining is not None and count >= remaining:
                    break

        # Add ellipsis if we couldn't show all children
        if limited and len(child_tasks) > tasks_to_show:
            parent_branch.add("[dim]...[/dim]")
            count += 1

        return count

    def _generate_task_table(self, max_rows: Optional[int] = None) -> Table:
        """Generate a table of tasks.

        Args:
            max_rows: Maximum number of task rows to display (not including header)
        """
        table = Table(box=box.SIMPLE, show_header=True, header_style="bold")
        table.add_column("ID", style="dim", width=8)
        table.add_column("Description")
        table.add_column("Status", justify="center")
        table.add_column("Added", style="dim")
        table.add_column("Completed", style="dim")

        # Determine if we need to limit rows
        total_tasks = len(self.tasks)
        limited = max_rows is not None and total_tasks > max_rows
        rows_to_show = max_rows if limited else total_tasks

        # Add task rows
        for i, (task_id, task) in enumerate(list(self.tasks.items())[:rows_to_show]):
            status_style = self._get_status_style(task["status"])
            added_time = task["time_added"].strftime("%H:%M:%S")
            completed_time = (
                task.get("time_completed", "").strftime("%H:%M:%S")
                if "time_completed" in task
                else "-"
            )

            table.add_row(
                f"{i+1:2d}",
                task["description"],
                status_style,
                added_time,
                completed_time,
            )

        # Add ellipsis row if we limited the rows
        if limited:
            table.add_row("...", "...", "...", "...", "...")

        return table

    def _get_level_style(self, level: str) -> str:
        """Get the appropriate style for a log level."""
        level = level.upper()
        if level == "INFO":
            return "bright_blue"
        elif level == "WARNING":
            return "yellow"
        elif level == "ERROR":
            return "red"
        elif level == "DEBUG":
            return "green"
        elif level == "CRITICAL":
            return "red bold"
        elif level == "SUCCESS":
            return "green bold"
        else:
            return "cyan"

    def _get_status_style(self, status: str) -> str:
        """Get styled status text based on status value."""
        if status == "completed":
            return "\N{WHITE HEAVY CHECK MARK} [green]COMPLETED[/green]"
        elif status == "pending":
            return "\N{HOURGLASS} [yellow]PENDING[/yellow]"
        elif status == "failed":
            return "\N{CROSS MARK} [red]FAILED[/red]"
        elif status == "running":
            current_spinner_char = SPINNER[int(time.time() * 4) % 4]
            return f"[blue]{current_spinner_char} RUNNING[/blue]"
        else:
            return "\N{INFORMATION SOURCE}" + f"[cyan]{status.upper()}[/cyan]"



"""
Retry decorator utility for handling transient failures with Rich visualization.
"""

import time
from functools import wraps
from typing import Type, List


def retry(
    max_attempts: int = 3,
    delay: float = 1.0,
    backoff_factor: float = 3.0,
    exceptions: List[Type[Exception]] = None,
):
    """
    Retry decorator for handling transient failures with Rich visualization.
    Uses techniques that don't conflict with other Live displays.

    Args:
        max_attempts: Maximum number of retry attempts (default: 3)
        delay: Initial delay between retries in seconds (default: 1.0)
        backoff_factor: Multiplier for delay between retries (default: 2.0)
        exceptions: List of exceptions to catch (default: all exceptions)

    Returns:
        A decorator function that will retry the decorated function on failure.
    """
    exceptions = exceptions or [Exception]

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            attempt = 1
            current_delay = delay

            while attempt <= max_attempts:
                try:
                    return func(*args, **kwargs)

                except tuple(exceptions) as e:
                    if attempt < max_attempts:
                        print(
                            f"Attempt {attempt}/{max_attempts}. Retrying in {current_delay:.2f}s\nError: {e}"
                        )
                        time.sleep(current_delay)
                        current_delay *= backoff_factor
                        attempt += 1
                    else:
                        print(
                            f"Attempt {attempt}/{max_attempts}. Max retries exceeded. Exiting.\nError: {e}"
                        )
                        time.sleep(1)
                        exit(1)

        return wrapper

    return decorator



"""
Pipeline module for orchestrating the flow of data through a series of agents.
"""

from typing import List, Optional
from agilemind.stage import Stage
from agilemind.context import Context
from dataclasses import dataclass, field


@dataclass
class Pipeline:
    """
    Pipeline class that orchestrates the flow between stages of tasks.
    """

    name: str
    description: Optional[str] = None
    stages: List[Stage] = field(default_factory=list)
    context: Context = field(default_factory=Context)

    def __post_init__(self):
        """Check that the pipeline has a name."""
        if not self.name:
            raise ValueError("Pipeline name is required")

    def register_executor(self, executor):
        self.context.executor = executor

    def add_stage(self, *stages: Stage) -> "Pipeline":
        """
        Add one or more stages to the pipeline.

        Args:
            *stages: One or more pipeline stages to add

        Returns:
            Self for chaining
        """
        self.stages.extend(stages)
        return self

    def run(self) -> Context:
        """
        Run the entire pipeline from start to finish.

        Args:
            initial_context: Initial context data to start with

        Returns:
            Final context after all stages have executed
        """
        for i, stage in enumerate(self.stages):
            try:
                self.context = stage.execute(self.context)
                self.context.set_metadata("last_executed_stage", stage.name)
            except Exception as e:
                self.context.set_metadata("failed_stage", stage.name)
                self.context.set_metadata("error", str(e))
                raise

        return self.context

    def run_until(self, stage_name: str) -> Context:
        """
        Run the pipeline until a specific stage is reached.

        Args:
            stage_name: Name of the stage to stop at (inclusive)
            initial_context: Initial context data to start with

        Returns:
            Context after executing up to and including the specified stage
        """
        for stage in self.stages:
            context_dict = self.context.to_dict()
            updated_context = stage.execute(context_dict)
            self.context.update(updated_context)
            self.context.set_metadata("last_executed_stage", stage.get_name())
            if stage.get_name() == stage_name:
                break

        return self.context
